#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
import re
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, cast

from jsonbender import F, Forall, If, K, OptionalS, S  # type: ignore

from otterdog.logging import get_logger
from otterdog.models import (
    EmbeddedModelObject,
    FailureType,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.utils import (
    UNSET,
    IndentingPrinter,
    associate_by_key,
    is_set_and_present,
    is_set_and_valid,
    is_unset,
    unwrap,
    write_patch_object_as_json,
)

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider

RS = TypeVar("RS", bound="Ruleset")

_logger = get_logger(__name__)


@dataclasses.dataclass
class PullRequestSettings(EmbeddedModelObject):
    required_approving_review_count: int
    dismisses_stale_reviews: bool
    requires_code_owner_review: bool
    requires_last_push_approval: bool
    requires_review_thread_resolution: bool

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        for key in self.keys(False):
            value = self.__getattribute__(key)
            if is_unset(value):
                context.add_failure(
                    FailureType.ERROR,
                    f"{parent_object.get_model_header(parent_object)} has not set required parameter "
                    f"'required_pull_request.{key}'.",
                )

        if is_set_and_valid(self.required_approving_review_count):
            if self.required_approving_review_count < 0 or self.required_approving_review_count > 10:
                context.add_failure(
                    FailureType.ERROR,
                    f"{parent_object.get_model_header(parent_object)} has "
                    f"'required_pull_request.required_approving_review_count' of value "
                    f"'{self.required_approving_review_count}' while only integers in the range [0, 10] are allowed.",
                )

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_pull_request}"

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "required_approving_review_count": OptionalS("required_approving_review_count", default=UNSET),
                "dismisses_stale_reviews": OptionalS("dismiss_stale_reviews_on_push", default=UNSET),
                "requires_code_owner_review": OptionalS("require_code_owner_review", default=UNSET),
                "requires_last_push_approval": OptionalS("require_last_push_approval", default=UNSET),
                "requires_review_thread_resolution": OptionalS("required_review_thread_resolution", default=UNSET),
            }
        )

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "dismiss_stale_reviews_on_push": S("dismisses_stale_reviews"),
                "require_code_owner_review": S("requires_code_owner_review"),
                "require_last_push_approval": S("requires_last_push_approval"),
                "required_review_thread_resolution": S("requires_review_thread_resolution"),
            }
        )

        for key in [
            "dismisses_stale_reviews",
            "requires_code_owner_review",
            "requires_last_push_approval",
            "requires_review_thread_resolution",
        ]:
            if key in mapping:
                mapping.pop(key)

        return mapping


@dataclasses.dataclass
class StatusCheckSettings(EmbeddedModelObject):
    do_not_enforce_on_create: bool
    strict: bool
    status_checks: list[str]

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        for key in ["strict", "status_checks"]:
            value = self.__getattribute__(key)
            if is_unset(value):
                context.add_failure(
                    FailureType.ERROR,
                    f"{parent_object.get_model_header(parent_object)} has not set required parameter "
                    f"'required_status_checks.{key}'.",
                )

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_status_checks}"

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        def transform_status_check(status_check):
            if "app_slug" in status_check:
                return status_check["app_slug"] + ":" + status_check["context"]
            else:
                return status_check["context"]

        mapping.update(
            {
                "do_not_enforce_on_create": OptionalS("do_not_enforce_on_create", default=UNSET),
                "strict": OptionalS("strict_required_status_checks_policy", default=UNSET),
                "status_checks": OptionalS("required_status_checks", default=[]) >> Forall(transform_status_check),
            }
        )

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        if "status_checks" in data:

            async def get_app_ids(status_checks) -> dict[str, str]:
                app_slugs = set()

                for check in status_checks:
                    if ":" in check:
                        app_slug, context = re.split(":", check, maxsplit=1)

                        if app_slug != "any" and " " not in app_slug:
                            app_slugs.add(app_slug)

                return await provider.get_app_ids(app_slugs)

            app_ids = await get_app_ids(data["status_checks"])
        else:
            app_ids = {}

        def transform_status_check(status_check):
            if ":" in status_check:
                app_slug, context = re.split(":", status_check, maxsplit=1)

                if app_slug == "any":
                    app_slug = None
                elif " " in app_slug:
                    context = f"{app_slug}:{context}"
                    app_slug = None
            else:
                app_slug = None
                context = status_check

            if app_slug is None:
                return {"context": context}
            else:
                return {"integration_id": app_ids[app_slug], "context": context}

        mapping.update(
            {
                "strict_required_status_checks_policy": S("strict"),
                "required_status_checks": S("status_checks") >> Forall(transform_status_check),
            }
        )

        for key in [
            "strict",
            "status_checks",
        ]:
            if key in mapping:
                mapping.pop(key)

        return mapping


@dataclasses.dataclass
class MergeQueueSettings(EmbeddedModelObject):
    merge_method: str
    build_concurrency: int
    min_group_size: int
    max_group_size: int
    wait_time_for_minimum_group_size: int
    status_check_timeout: int
    requires_all_group_entries_to_pass_required_checks: bool

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        for key in self.keys(False):
            value = self.__getattribute__(key)
            if is_unset(value):
                context.add_failure(
                    FailureType.ERROR,
                    f"{parent_object.get_model_header(parent_object)} has not set required parameter "
                    f"'required_merge_queue.{key}'.",
                )

        if self.merge_method not in {"MERGE", "SQUASH", "REBASE"}:
            context.add_failure(
                FailureType.ERROR,
                f"{parent_object.get_model_header(parent_object)} has 'required_merge_queue.merge_method' of value "
                f"'{self.merge_method}', only values ('MERGE' | 'SQUASH' | 'REBASE') are allowed.",
            )

        for key in [
            "build_concurrency",
            "min_group_size",
            "max_group_size",
            "wait_time_for_minimum_group_size",
            "status_check_timeout",
        ]:
            value = self.__getattribute__(key)
            if is_set_and_valid(value):
                if int(value) < 0:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{parent_object.get_model_header(parent_object)} has 'required_merge_queue.{key}' of value "
                        f"'{value}' while only non-negative integers are allowed.",
                    )

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_merge_queue}"

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "build_concurrency": S("max_entries_to_build"),
                "status_check_timeout": S("check_response_timeout_minutes"),
                "min_group_size": S("min_entries_to_merge"),
                "max_group_size": S("max_entries_to_merge"),
                "wait_time_for_minimum_group_size": S("min_entries_to_merge_wait_minutes"),
                "requires_all_group_entries_to_pass_required_checks": If(
                    S("grouping_strategy") == K("ALLGREEN"), K(True), K(False)
                ),
            }
        )

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "max_entries_to_build": S("build_concurrency"),
                "check_response_timeout_minutes": S("status_check_timeout"),
                "min_entries_to_merge": S("min_group_size"),
                "max_entries_to_merge": S("max_group_size"),
                "min_entries_to_merge_wait_minutes": S("wait_time_for_minimum_group_size"),
                "grouping_strategy": If(
                    S("requires_all_group_entries_to_pass_required_checks") == K(True), K("ALLGREEN"), K("HEADGREEN")
                ),
            }
        )

        for key in [
            "build_concurrency",
            "status_check_timeout",
            "min_group_size",
            "max_group_size",
            "wait_time_for_minimum_group_size",
            "requires_all_group_entries_to_pass_required_checks",
        ]:
            if key in mapping:
                mapping.pop(key)

        return mapping


@dataclasses.dataclass
class Ruleset(ModelObject, abc.ABC):
    """
    Represents a Ruleset.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    node_id: str = dataclasses.field(metadata={"external_only": True})
    target: str
    enforcement: str

    bypass_actors: list[str]

    include_refs: list[str]
    exclude_refs: list[str]

    allows_creations: bool
    allows_deletions: bool
    allows_updates: bool
    allows_force_pushes: bool

    requires_commit_signatures: bool
    requires_linear_history: bool

    requires_deployments: bool
    required_deployment_environments: list[str]

    required_pull_request: PullRequestSettings | None = dataclasses.field(metadata={"embedded_model": True})
    required_status_checks: StatusCheckSettings | None = dataclasses.field(metadata={"embedded_model": True})
    required_merge_queue: MergeQueueSettings | None = dataclasses.field(metadata={"embedded_model": True})

    _roles: ClassVar[dict[str, str]] = {"5": "RepositoryAdmin", "4": "Write", "2": "Maintain", "1": "OrganizationAdmin"}
    _inverted_roles: ClassVar[dict[str, str]] = {v: k for k, v in _roles.items()}

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        from .github_organization import GitHubOrganization

        org_settings = cast(GitHubOrganization, context.root_object).settings

        if is_set_and_valid(self.target):
            if self.target not in {"branch", "tag", "push"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'target' of value "
                    f"'{self.target}', while only values ('branch' | 'tag' | 'push') are allowed.",
                )

        enterprise_plan = org_settings.plan == "enterprise"
        if is_set_and_valid(self.enforcement):
            if self.enforcement not in {"active", "disabled", "evaluate"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'enforcement' of value '{self.enforcement}', "
                    f"while only values ('active' | 'disabled' | 'evaluate') are allowed.",
                )

            if enterprise_plan is False and self.enforcement == "evaluate":
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'enforcement' of value "
                    f"'{self.enforcement}' which is only available for an 'enterprise' plan.",
                )

        def valid_branch_condition_pattern(pattern: str) -> bool:
            return pattern.startswith("refs/heads/") or pattern == "~DEFAULT_BRANCH" or pattern == "~ALL"

        def valid_tag_condition_pattern(pattern: str) -> bool:
            return pattern.startswith("refs/tags/") or pattern == "~ALL"

        if is_set_and_valid(self.include_refs):
            for ref in self.include_refs:
                if (self.target == "branch" or self.target == "push") and not valid_branch_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'include_refs' pattern "
                        f"'{ref}', while only values ('refs/heads/*', '~DEFAULT_BRANCH', '~ALL') are allowed",
                    )
                elif self.target == "tag" and not valid_tag_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'include_refs' pattern "
                        f"'{ref}', while only values ('refs/tags/*', '~ALL') are allowed",
                    )

        if is_set_and_valid(self.exclude_refs):
            for ref in self.exclude_refs:
                if (self.target == "branch" or self.target == "push") and not valid_branch_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'exclude_refs' pattern "
                        f"'{ref}', while only values ('refs/heads/*', '~DEFAULT_BRANCH', '~ALL') are allowed",
                    )
                elif self.target == "tag" and not valid_tag_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'exclude_refs' pattern "
                        f"'{ref}', while only values ('refs/tags/*', '~ALL') are allowed",
                    )

        # if 'requires_deployments' is disabled, issue a warning if required_deployment_environments is non-empty.
        if (
            self.requires_deployments is False
            and is_set_and_valid(self.required_deployment_environments)
            and len(self.required_deployment_environments) > 0
        ):
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has "
                f"'requires_deployments' disabled but 'required_deployment_environments' is set to "
                f"'{self.required_deployment_environments}', setting will be ignored.",
            )

        if self.requires_deployments is True and len(self.required_deployment_environments) > 0:
            from .repository import Repository

            environments = cast(Repository, parent_object).environments

            environments_by_name = associate_by_key(environments, lambda x: x.name)
            for env_name in self.required_deployment_environments:
                if env_name not in environments_by_name:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} requires deployment environment "
                        f"'{env_name}' which is not defined in the repository itself.",
                    )

        if is_set_and_present(self.required_pull_request):
            self.required_pull_request.validate(context, parent_object)

        if is_set_and_present(self.required_merge_queue):
            self.required_merge_queue.validate(context, parent_object)

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.requires_deployments is False:
            if field.name in ["required_deployment_environments"]:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    @classmethod
    def get_mapping_from_model(cls) -> dict[str, Any]:
        mapping = super().get_mapping_from_model()

        mapping.update(
            {
                "required_pull_request": If(
                    OptionalS("required_pull_request", default=None) == K(None),
                    K(None),
                    S("required_pull_request") >> F(lambda x: PullRequestSettings.from_model_data(x)),
                ),
                "required_status_checks": If(
                    OptionalS("required_status_checks", default=None) == K(None),
                    K(None),
                    S("required_status_checks") >> F(lambda x: StatusCheckSettings.from_model_data(x)),
                ),
                "required_merge_queue": If(
                    OptionalS("required_merge_queue", default=None) == K(None),
                    K(None),
                    S("required_merge_queue") >> F(lambda x: MergeQueueSettings.from_model_data(x)),
                ),
            }
        )

        return mapping

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping.update(
            {
                "include_refs": OptionalS("conditions", "ref_name", "include", default=[]),
                "exclude_refs": OptionalS("conditions", "ref_name", "exclude", default=[]),
            }
        )

        rules = data.get("rules", [])

        def check_simple_rule(prop_key: str, rule_type: str, value_if_rule_is_present: bool) -> None:
            if any((_ := rule) for rule in rules if rule["type"] == rule_type):
                mapping[prop_key] = K(value_if_rule_is_present)
            else:
                mapping[prop_key] = K(not value_if_rule_is_present)

        check_simple_rule("allows_deletions", "deletion", False)
        check_simple_rule("allows_creations", "creation", False)
        check_simple_rule("allows_updates", "update", False)
        check_simple_rule("allows_force_pushes", "non_fast_forward", False)

        check_simple_rule("requires_commit_signatures", "required_signatures", True)
        check_simple_rule("requires_linear_history", "required_linear_history", True)

        # bypass actors
        if "bypass_actors" in data:
            bypass_actors = data["bypass_actors"]
            transformed_actors = []
            for actor in bypass_actors:
                actor_type = actor["actor_type"]

                if actor_type == "RepositoryRole":
                    role = cls._roles.get(str(actor["actor_id"]), None)
                    if role is None:
                        _logger.warning("fail to map repository role '%s', skipping", actor["actor_id"])
                        continue
                    transformed_actor = f"#{role}"
                elif actor_type == "OrganizationAdmin":
                    transformed_actor = "#OrganizationAdmin"
                elif actor_type == "Team":
                    transformed_actor = f"@{actor['team_slug']}"
                elif actor_type == "Integration":
                    transformed_actor = actor["app_slug"]
                else:
                    continue

                bypass_mode = actor.get("bypass_mode", "always")
                if bypass_mode != "always":
                    transformed_actor = f"{transformed_actor}:{bypass_mode}"

                transformed_actors.append(transformed_actor)

            mapping["bypass_actors"] = K(transformed_actors)

        else:
            mapping["bypass_actors"] = K([])

        # required pull request
        if any((found := rule) for rule in rules if rule["type"] == "pull_request"):
            parameters = found.get("parameters", {})
            mapping["required_pull_request"] = K(PullRequestSettings.from_provider_data(org_id, parameters))
        else:
            mapping["required_pull_request"] = K(None)

        # required status checks
        if any((found := rule) for rule in rules if rule["type"] == "required_status_checks"):
            parameters = found.get("parameters", {})
            mapping["required_status_checks"] = K(StatusCheckSettings.from_provider_data(org_id, parameters))
        else:
            mapping["required_status_checks"] = K(None)

        # required deployments
        if any((found := rule) for rule in rules if rule["type"] == "required_deployments"):
            mapping["requires_deployments"] = K(True)
            deployment_environments = found.get("parameters", {}).get("required_deployment_environments", [])
            mapping["required_deployment_environments"] = K(deployment_environments)
        else:
            mapping["requires_deployments"] = K(False)
            mapping["required_deployment_environments"] = K([])

        # required merge queue
        if any((found := rule) for rule in rules if rule["type"] == "merge_queue"):
            merge_queue_parameters = found.get("parameters", {})
            mapping["required_merge_queue"] = K(MergeQueueSettings.from_provider_data(org_id, merge_queue_parameters))
        else:
            mapping["required_merge_queue"] = K(None)

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        def pop_mapping(keys: list[str]) -> None:
            for key_to_pop in keys:
                if key_to_pop in mapping:
                    mapping.pop(key_to_pop)

        # include / excludes
        ref_names = {}
        if "include_refs" in data:
            mapping.pop("include_refs")
            ref_names["include"] = K(data["include_refs"])

        if "exclude_refs" in data:
            mapping.pop("exclude_refs")
            ref_names["exclude"] = K(data["exclude_refs"])

        if len(ref_names) > 0:
            mapping["conditions"] = {"ref_name": ref_names}

        # bypass actors
        if "bypass_actors" in data:
            bypass_actors: list[str] = data["bypass_actors"]
            transformed_actors = []

            def extract_actor_and_bypass_mode(encoded_data: str) -> tuple[str, str]:
                if ":" in encoded_data:
                    a, m = re.split(":", encoded_data, maxsplit=1)
                else:
                    a = encoded_data
                    m = "always"

                return a, m

            for actor in bypass_actors:
                if actor.startswith("#"):
                    role, bypass_mode = extract_actor_and_bypass_mode(actor[1:])
                    actor_type = "RepositoryRole"
                    actor_id = cls._inverted_roles[role]
                    if actor_id == "1":
                        actor_type = "OrganizationAdmin"
                elif actor.startswith("@"):
                    team, bypass_mode = extract_actor_and_bypass_mode(actor[1:])
                    actor_type = "Team"
                    actor_id = (await provider.rest_api.team.get_team_ids(team))[0]
                else:
                    app, bypass_mode = extract_actor_and_bypass_mode(actor)
                    actor_type = "Integration"
                    actor_id = (await provider.rest_api.app.get_app_ids(app))[0]

                transformed_actors.append(
                    {"actor_id": K(int(actor_id)), "actor_type": K(actor_type), "bypass_mode": K(bypass_mode)}
                )

            mapping["bypass_actors"] = transformed_actors

        # rules
        rules: list[Any] = []

        def add_simple_rule(prop_key: str, rule_type: str, inverted: bool) -> None:
            if prop_key in data:
                mapping.pop(prop_key)
                prop_value = data[prop_key]

                if (inverted and not prop_value) or (not inverted and prop_value):
                    rules.append({"type": K(rule_type)})

        add_simple_rule("allows_deletions", "deletion", True)
        add_simple_rule("allows_creations", "creation", True)
        add_simple_rule("allows_updates", "update", True)
        add_simple_rule("allows_force_pushes", "non_fast_forward", True)

        add_simple_rule("requires_commit_signatures", "required_signatures", False)
        add_simple_rule("requires_linear_history", "required_linear_history", False)

        def add_parameter(prop_key: str, param_key: str, params: dict[str, Any]):
            if prop_key in data:
                params[param_key] = S(prop_key)

        # required pull request
        if "required_pull_request" in data:
            mapping.pop("required_pull_request")
            required_pull_request = data["required_pull_request"]
            if required_pull_request is not None:
                pull_request_parameters = await PullRequestSettings.dict_to_provider_data(
                    org_id,
                    required_pull_request,
                    provider,
                )
                if pull_request_parameters and len(pull_request_parameters) > 0:
                    rule = {
                        "type": K("pull_request"),
                        "parameters": K(pull_request_parameters),
                    }
                    rules.append(rule)

        # required status checks
        if "required_status_checks" in data:
            mapping.pop("required_status_checks")
            required_status_checks = data["required_status_checks"]
            if required_status_checks is not None:
                status_check_parameters = await StatusCheckSettings.dict_to_provider_data(
                    org_id,
                    required_status_checks,
                    provider,
                )
                if status_check_parameters and len(status_check_parameters) > 0:
                    rule = {
                        "type": K("required_status_checks"),
                        "parameters": K(status_check_parameters),
                    }
                    rules.append(rule)

        # required deployments
        if "requires_deployments" in data:
            value = data["requires_deployments"]
            if value is True:
                rule = {"type": K("required_deployments")}
                deployment_parameters: dict[str, Any] = {}
                add_parameter(
                    "required_deployment_environments", "required_deployment_environments", deployment_parameters
                )
                rule["parameters"] = deployment_parameters
                rules.append(rule)

        pop_mapping(["requires_deployments", "required_deployment_environments"])

        # required merge queue
        if "required_merge_queue" in data:
            mapping.pop("required_merge_queue")
            required_merge_queue = data["required_merge_queue"]
            if required_merge_queue is not None:
                merge_queue_parameters = await MergeQueueSettings.dict_to_provider_data(
                    org_id,
                    required_merge_queue,
                    provider,
                )
                if merge_queue_parameters and len(merge_queue_parameters) > 0:
                    rule = {
                        "type": K("merge_queue"),
                        "parameters": K(merge_queue_parameters),
                    }
                    rules.append(rule)

        if len(rules) > 0:
            mapping["rules"] = rules

        return mapping

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)

        template_function = self.get_jsonnet_template_function(jsonnet_config, False)

        if "name" in patch:
            patch.pop("name")

        printer.print(f"{unwrap(template_function)}('{self.name}')")

        if "required_pull_request" in patch and patch.get("required_pull_request") is not None:
            patch.pop("required_pull_request")

        if "required_merge_queue" in patch and patch.get("required_merge_queue") is not None:
            patch.pop("required_merge_queue")

        if "required_status_checks" in patch and patch.get("required_status_checks") is not None:
            patch.pop("required_status_checks")

        write_patch_object_as_json(patch, printer, close_object=False)

        if is_set_and_present(self.required_pull_request):
            default_pull_request_config = cast(Ruleset, default_object).required_pull_request
            if default_pull_request_config is None:
                default_pull_request_config = PullRequestSettings.from_model_data(
                    jsonnet_config.default_pull_request_config
                )
                embedded_extend = False
            else:
                embedded_extend = True

            if is_set_and_valid(default_pull_request_config):
                patch = self.required_pull_request.get_patch_to(default_pull_request_config)
                if len(patch) > 0:
                    printer.print(f"required_pull_request{'+' if embedded_extend else ''}:")
                    self.required_pull_request.to_jsonnet(
                        printer,
                        jsonnet_config,
                        context,
                        embedded_extend,
                        default_pull_request_config,
                    )

        if is_set_and_present(self.required_merge_queue):
            default_merge_queue_config = cast(Ruleset, default_object).required_merge_queue
            if default_merge_queue_config is None:
                default_merge_queue_config = MergeQueueSettings.from_model_data(
                    jsonnet_config.default_merge_queue_config
                )
                embedded_extend = False
            else:
                embedded_extend = True

            if is_set_and_valid(default_merge_queue_config):
                patch = self.required_merge_queue.get_patch_to(default_merge_queue_config)
                if len(patch) > 0:
                    printer.print(f"required_merge_queue{'+' if embedded_extend else ''}:")
                    self.required_merge_queue.to_jsonnet(
                        printer,
                        jsonnet_config,
                        context,
                        embedded_extend,
                        default_merge_queue_config,
                    )

        if is_set_and_present(self.required_status_checks):
            default_status_check_config = cast(Ruleset, default_object).required_status_checks
            if default_status_check_config is None:
                default_status_check_config = StatusCheckSettings.from_model_data(
                    jsonnet_config.default_status_checks_config
                )
                embedded_extend = False
            else:
                embedded_extend = True

            if is_set_and_valid(default_status_check_config):
                patch = self.required_status_checks.get_patch_to(default_status_check_config)
                if len(patch) > 0:
                    printer.print(f"required_status_checks{'+' if embedded_extend else ''}:")
                    self.required_status_checks.to_jsonnet(
                        printer,
                        jsonnet_config,
                        context,
                        embedded_extend,
                        default_status_check_config,
                    )

        # close the object
        printer.level_down()
        printer.println("},")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from aiofiles import open, os, ospath
from aioshutil import copy

from otterdog.models import PatchContext
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from collections.abc import Sequence

    from otterdog.config import OrganizationConfig
    from otterdog.models.webhook import Webhook


class ImportOperation(Operation):
    """
    Imports the current live configuration for organizations.
    """

    def __init__(self, force_processing: bool, no_web_ui: bool):
        super().__init__()
        self._force_processing = force_processing
        self._no_web_ui = no_web_ui

    @property
    def force_processing(self) -> bool:
        return self._force_processing

    @property
    def no_web_ui(self) -> bool:
        return self._no_web_ui

    def pre_execute(self) -> None:
        self.printer.println("Importing resources:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()

        self._print_project_header(org_config, org_index, org_count)

        org_file_name = jsonnet_config.org_config_file
        if not await self.check_config_file_overwrite_if_exists(org_file_name, self.force_processing):
            return 1

        if await ospath.exists(org_file_name):
            sync_from_previous_config = True
            backup_file = f"{org_file_name}.bak"
            await copy(org_file_name, backup_file)
            self.printer.println(f"\nExisting definition copied to '[bold]{backup_file}[/]'.\n")
        else:
            sync_from_previous_config = False

        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            if self.no_web_ui is True:
                self.printer.print_warn(
                    "the Web UI will not be queried as '--no-web-ui' has been specified, "
                    "the resulting config will be incomplete."
                )

            async with GitHubProvider(credentials) as provider:
                organization = await GitHubOrganization.load_from_provider(
                    org_config.name,
                    github_id,
                    jsonnet_config,
                    provider,
                    self.no_web_ui,
                    exclude_teams=self.config.exclude_teams_pattern,
                )

            # copy secrets from existing configuration if it is present.
            if sync_from_previous_config:
                previous_organization = GitHubOrganization.load_from_file(github_id, org_file_name)

                self.printer.println("Copying secrets from previous configuration.")
                organization.copy_secrets(previous_organization)

                self.printer.print("Masking webhooks from previous configuration... ")
                masked_urls = 0
                for org_webhook in previous_organization.webhooks:
                    if org_webhook.url.endswith("*"):
                        masked_urls += _mask_webhook_url(organization.webhooks, org_webhook)

                for repo in previous_organization.repositories:
                    for repo_webhook in repo.webhooks:
                        if repo_webhook.url.endswith("*"):
                            new_repo = organization.get_repository(repo.name)
                            if new_repo is not None:
                                masked_urls += _mask_webhook_url(new_repo.webhooks, repo_webhook)

                self.printer.println(f"{masked_urls} URLs have been masked.")

            context = PatchContext(github_id, organization.settings)
            output = organization.to_jsonnet(jsonnet_config, context)

            output_dir = jsonnet_config.org_dir
            if not await ospath.exists(output_dir):
                await os.makedirs(output_dir)

            async with open(org_file_name, "w") as file:
                await file.write(output)

            self.printer.println(f"Organization definition written to '{org_file_name}'.")

            return 0
        finally:
            self.printer.level_down()


def _mask_webhook_url(webhooks: Sequence[Webhook], masked_webhook: Webhook) -> int:
    stripped_url = masked_webhook.url.rstrip("*")
    for webhook in webhooks:
        if webhook.url.startswith(stripped_url):
            webhook.url = masked_webhook.url
            return 1

    return 0



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.team import Team
from otterdog.utils import UNSET, query_json

from . import ModelTest


class TeamTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return Team.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_org_team

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-team.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-team.json")

    def test_load_from_model(self):
        team = Team.from_model_data(self.model_data)

        assert team.id is UNSET
        assert team.slug is UNSET
        assert team.name == "Test Team"
        assert team.description == "Blabla"
        assert team.privacy == "visible"
        assert team.notifications is True
        assert team.members == ["netomi"]

    def test_load_from_provider(self):
        team = Team.from_provider_data(self.org_id, self.provider_data)

        assert team.id == 1102681190
        assert team.slug == "test-team"
        assert team.name == "Test Team"
        assert team.description == "Blabla"
        assert team.privacy == "visible"
        assert team.notifications is True
        assert team.members == ["netomi"]

    async def test_to_provider(self):
        team = Team.from_model_data(self.model_data)

        provider_data = await team.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 5
        assert provider_data["name"] == "Test Team"
        assert provider_data["privacy"] == "closed"
        assert provider_data["notification_setting"] == "notifications_enabled"

        assert query_json("members[0]", provider_data) == "netomi"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class CheckTokenPermissionsOperation(Operation):
    """
    Checks the granted permissions for the otterdog-token.
    """

    EXPECTED_SCOPES = frozenset(["admin:org", "admin:org_hook", "delete_repo", "repo", "workflow"])

    def __init__(self, list_granted_scopes: bool):
        super().__init__()
        self._list_granted_scopes = list_granted_scopes

    @property
    def list_granted_scopes(self) -> bool:
        return self._list_granted_scopes

    def pre_execute(self) -> None:
        self.printer.println("Checking token permissions:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                scopes = await provider.rest_api.meta.get_scopes()

                if self._list_granted_scopes is True:
                    self.printer.println(f"[green]Granted scopes[/]: {scopes}")

                missing_scopes = self._get_missing_scopes(scopes)
                if len(missing_scopes) > 0:
                    scope_string = ", ".join(missing_scopes)
                    self.printer.println(f"[red]Missing scopes[/]: {scope_string}")
                    return 1

            return 0
        finally:
            self.printer.level_down()

    def _get_missing_scopes(self, scopes: str) -> list[str]:
        missing_scopes = []
        granted_scopes = {s.strip() for s in scopes.split(",") if len(s.strip()) > 0}
        for scope in self.EXPECTED_SCOPES:
            if scope not in granted_scopes:
                missing_scopes.append(scope)

        return missing_scopes



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from dataclasses import fields
from unittest import mock

import yaml

from otterdog.models.repository import Repository
from otterdog.webapp.blueprints import BlueprintType, read_blueprint
from otterdog.webapp.blueprints.required_file import RequiredFileBlueprint

yaml_content = """
id: require-default-security-md
name: Requires SECURITY.md file
type: required_file
config:
  repo_selector:
    name_pattern: .github
  files:
    - path: SECURITY.md
      content: |
        This is our security policy.
        Please head over to....
"""

multiple_repo_selection = """
id: require-yml-file
name: Require file
type: required_file
config:
    repo_selector:
      name_pattern:
        - .github
        - repo-1
        - repo-2
        - repo-3
    files:
      - path: .github/workflows/dependabot-auto-merge.yml
        content: |
          blabla
"""

no_repo_selection = """
id: require-yml-file
name: Require file
type: required_file
config:
    files:
      - path: .github/workflows/dependabot-auto-merge.yml
        content: |
          blabla
"""


def test_read():
    config = yaml.safe_load(yaml_content)

    blueprint = read_blueprint("a", config)
    assert blueprint.type == BlueprintType.REQUIRED_FILE

    assert isinstance(blueprint, RequiredFileBlueprint)


def test_repo_selector_multiple_repos():
    config = yaml.safe_load(multiple_repo_selection)

    blueprint = read_blueprint("a", config)
    assert blueprint.type == BlueprintType.REQUIRED_FILE

    assert isinstance(blueprint, RequiredFileBlueprint)
    assert len(blueprint.files) == 1

    required_file = blueprint.files[0]

    assert required_file.strict is False

    selector = blueprint.repo_selector

    assert selector.matches(create_repo_with_name(".github"))
    assert selector.matches(create_repo_with_name("repo-1"))
    assert selector.matches(create_repo_with_name("repo-10")) is False


def test_no_repo_selector():
    config = yaml.safe_load(no_repo_selection)

    blueprint = read_blueprint("a", config)
    assert blueprint.type == BlueprintType.REQUIRED_FILE

    assert isinstance(blueprint, RequiredFileBlueprint)
    assert len(blueprint.files) == 1

    required_file = blueprint.files[0]

    assert required_file.strict is False

    selector = blueprint.repo_selector

    assert selector is None

    assert blueprint._matches(create_repo_with_name(".github"))
    assert blueprint._matches(create_repo_with_name("repo-1"))
    assert blueprint._matches(create_repo_with_name("repo-10"))


def create_repo_with_name(name: str) -> Repository:
    repo = create_dataclass_mock(Repository)
    repo.name = name
    return repo


def create_dataclass_mock(obj):
    return mock.Mock(spec_set=[field.name for field in fields(obj)])



#  *******************************************************************************
#  Copyright (c) 2024-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************
import json
from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException

from . import RestApi, RestClient

_logger = get_logger(__name__)


class PullRequestClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_pull_request(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
    ) -> dict[str, Any]:
        _logger.debug("getting pull request with number '%s' from repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving pull request:\n{ex}") from ex

    async def create_pull_request(
        self,
        org_id: str,
        repo_name: str,
        title: str,
        head: str,
        base: str,
        body: str | None = None,
    ) -> dict[str, Any]:
        _logger.debug("creating pull request for repo '%s/%s'", org_id, repo_name)

        try:
            data = {
                "title": title,
                "head": head,
                "base": base,
            }

            if body is not None:
                data["body"] = body

            return await self.requester.request_json("POST", f"/repos/{org_id}/{repo_name}/pulls", data=data)
        except GitHubException as ex:
            raise RuntimeError(f"failed creating pull request:\n{ex}") from ex

    async def get_pull_requests(
        self,
        org_id: str,
        repo_name: str,
        state: str = "all",
        base_ref: str | None = None,
    ) -> list[dict[str, Any]]:
        _logger.debug("getting pull requests from repo '%s/%s'", org_id, repo_name)

        try:
            params = {"state": state}

            if base_ref is not None:
                params.update({"base": base_ref})

            return await self.requester.request_paged_json("GET", f"/repos/{org_id}/{repo_name}/pulls", params=params)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving pull requests:\n{ex}") from ex

    async def merge_pull_request(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
        method: str,
    ) -> dict[str, Any]:
        _logger.debug("merging pull request #%s for repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            data = {"merge_method": method}
            return await self.requester.request_json(
                "PUT", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/merge", data=data
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed merging pull request:\n{ex}") from ex

    async def get_commits(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
    ) -> list[dict[str, Any]]:
        _logger.debug("getting commits for pull request #%s from repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            return await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/commits"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving pull request commits:\n{ex}") from ex

    async def get_reviews(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
    ) -> list[dict[str, Any]]:
        _logger.debug("getting reviews for pull request #%s from repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            return await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/reviews"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving pull request reviews:\n{ex}") from ex

    async def request_reviews(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
        reviewers: list[str] | None = None,
        team_reviewers: list[str] | None = None,
    ) -> bool:
        _logger.debug(
            "requesting reviews for pull request #%s in repo '%s/%s': %s, %s",
            pull_request_number,
            org_id,
            repo_name,
            reviewers,
            team_reviewers,
        )

        if (reviewers is None or len(reviewers) == 0) and (team_reviewers is None or len(team_reviewers) == 0):
            _logger.error(
                "requesting reviews for pull request #%s in repo '%s/%s' without any reviewer specified",
                pull_request_number,
                org_id,
                repo_name,
            )
            return False

        try:
            data = {}

            if reviewers is not None:
                data["reviewers"] = reviewers

            if team_reviewers is not None:
                data["team_reviewers"] = team_reviewers

            status, body = await self.requester.request_raw(
                "POST",
                f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/requested_reviewers",
                data=json.dumps(data),
            )

            if status == 201:
                return True
            elif status == 422:
                _logger.warning("failed to request reviews for reviewers (%s, %s): %s", reviewers, team_reviewers, body)
                return False
            else:
                raise RuntimeError(
                    f"failed requesting reviews for pull request #{pull_request_number} in repo "
                    f"'{org_id}/{repo_name}'\n{status}: {body}"
                )

        except GitHubException as ex:
            raise RuntimeError(f"failed requesting pull request reviews:\n{ex}") from ex

    async def get_files(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
    ) -> list[dict[str, Any]]:
        _logger.debug("getting files for pull request #%s from repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            return await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/files"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving pull request files:\n{ex}") from ex

    async def merge(
        self,
        org_id: str,
        repo_name: str,
        pull_request_number: str,
        commit_message: str | None = None,
        merge_method: str = "squash",
    ) -> bool:
        _logger.debug("merging pull request #%s from repo '%s/%s'", pull_request_number, org_id, repo_name)

        try:
            data = {
                "merge_method": merge_method,
            }

            if commit_message is not None:
                data.update({"commit_message": commit_message})

            response = await self.requester.request_json(
                "PUT",
                f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}/merge",
                data=data,
            )
            return response["merged"]
        except GitHubException as ex:
            raise RuntimeError(f"failed merging pull request:\n{ex}") from ex



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException

from . import RestApi, RestClient

_logger = get_logger(__name__)


class ReferenceClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_branch_reference(self, org_id: str, repo_name: str, branch_name: str) -> dict[str, Any]:
        _logger.debug("getting branch reference with name '%s' from repo '%s/%s'", branch_name, org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/git/ref/heads/{branch_name}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving reference:\n{ex}") from ex

    async def get_matching_references(self, org_id: str, repo_name: str, ref_pattern: str) -> list[dict[str, Any]]:
        _logger.debug("getting matching references with pattern '%s' from repo '%s/%s'", ref_pattern, org_id, repo_name)

        try:
            return await self.requester.request_json(
                "GET",
                f"/repos/{org_id}/{repo_name}/git/matching-refs/{ref_pattern}",
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving matching references for repo '{org_id}/{repo_name}':\n{ex}") from ex

    async def create_reference(self, org_id: str, repo_name: str, ref: str, sha: str) -> str:
        _logger.debug("creating reference with name '%s' and sha '%s' for repo '%s/%s'", ref, sha, org_id, repo_name)

        try:
            data = {
                "ref": f"refs/heads/{ref}",
                "sha": sha,
            }

            await self.requester.request_json("POST", f"/repos/{org_id}/{repo_name}/git/refs", data=data)
            return data["ref"]
        except GitHubException as ex:
            raise RuntimeError(f"failed creating reference:\n{ex}") from ex

    async def delete_reference(self, org_id: str, repo_name: str, ref: str) -> bool:
        _logger.debug("deleting reference with name '%s' in repo '%s/%s'", ref, org_id, repo_name)

        full_ref = f"refs/heads/{ref}"
        status, body = await self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/git/{full_ref}")
        if status == 204:
            return True
        elif status in (409, 422):
            return False
        else:
            raise RuntimeError(
                f"failed deleting reference '{ref}' in repo '{org_id}/{repo_name}'" f"\n{status}: {body}"
            )



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.organization_secret import OrganizationSecret
from otterdog.utils import Change

from . import ModelTest


class OrganizationSecretTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return OrganizationSecret.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_org_secret

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-org-secret.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-org-secret.json")

    def test_load_from_model(self):
        secret = OrganizationSecret.from_model_data(self.model_data)

        assert secret.name == "TEST-SECRET"
        assert secret.visibility == "selected"
        assert secret.selected_repositories == ["test-repo"]
        assert secret.value == "1234"

    def test_load_from_provider(self):
        secret = OrganizationSecret.from_provider_data(self.org_id, self.provider_data)

        assert secret.name == "TEST-SECRET"
        assert secret.visibility == "selected"
        assert secret.selected_repositories == ["test-repo"]
        assert secret.value == "********"

    def test_patch(self):
        current = OrganizationSecret.from_model_data(self.model_data)
        default = OrganizationSecret.from_model_data(self.model_data)

        default.visibility = "public"

        patch = current.get_patch_to(default)

        assert len(patch) == 1
        assert patch["visibility"] == current.visibility

    def test_difference(self):
        current = OrganizationSecret.from_model_data(self.model_data)
        other = OrganizationSecret.from_model_data(self.model_data)

        other.visibility = "public"

        diff = current.get_difference_from(other)

        assert len(diff) == 1
        assert diff["visibility"] == Change(other.visibility, current.visibility)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from . import CacheStrategy

if TYPE_CHECKING:
    from typing import Any


def ghproxy_cache(uri: str) -> CacheStrategy:
    return _GHProxy(uri)


class _GHProxy(CacheStrategy):
    def __init__(self, proxy_uri: str):
        from urllib import parse

        self._proxy_uri = parse.urlparse(proxy_uri)

    def get_cache_backend(self) -> None:
        return None

    def is_external(self) -> bool:
        return True

    def replace_base_url(self, base_url: str) -> str:
        from urllib import parse

        url = parse.urlparse(base_url)
        return parse.urlunparse(self._proxy_uri._replace(path=url.path))

    def get_request_parameters(self) -> dict[str, Any]:
        return {}

    def __str__(self):
        return f"ghproxy-cache('{self._proxy_uri}')"



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import json
import os
import re
from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any, Protocol

from otterdog.credentials import CredentialProvider

from .jsonnet import JsonnetConfig
from .logging import get_logger
from .utils import deep_merge_dict, query_json

if TYPE_CHECKING:
    from collections.abc import Mapping

    from otterdog.credentials import Credentials

_logger = get_logger(__name__)


class OrganizationConfig:
    def __init__(
        self,
        name: str,
        github_id: str,
        config_repo: str,
        base_template: str,
        jsonnet_config: JsonnetConfig,
        credential_data: dict[str, Any],
    ):
        self._name = name
        self._github_id = github_id
        self._config_repo = config_repo
        self._base_template = base_template
        self._jsonnet_config = jsonnet_config
        self._credential_data = credential_data

    @property
    def name(self):
        return self._name

    @property
    def github_id(self) -> str:
        return self._github_id

    @property
    def config_repo(self) -> str:
        return self._config_repo

    @property
    def base_template(self) -> str:
        return self._base_template

    @property
    def jsonnet_config(self) -> JsonnetConfig:
        return self._jsonnet_config

    @property
    def credential_data(self) -> dict[str, Any]:
        return self._credential_data

    @credential_data.setter
    def credential_data(self, data: dict[str, Any]) -> None:
        self._credential_data = data

    def __repr__(self) -> str:
        return (
            f"OrganizationConfig('{self.name}', '{self.github_id}', "
            f"'{self.config_repo}', {json.dumps(self.credential_data)})"
        )

    @classmethod
    def from_dict(cls, data: dict[str, Any], otterdog_config: OtterdogConfig) -> OrganizationConfig:
        name = data.get("name")
        if name is None:
            raise RuntimeError(f"missing required name for organization config with data: '{json.dumps(data)}'")

        github_id = data.get("github_id")
        if github_id is None:
            raise RuntimeError(f"missing required github_id for organization config with name '{name}'")

        config_repo = data.get("config_repo", otterdog_config.default_config_repo)
        base_template = data.get("base_template", otterdog_config.default_base_template)

        jsonnet_config = JsonnetConfig(
            github_id,
            otterdog_config.jsonnet_base_dir,
            base_template,
            otterdog_config.local_mode,
        )

        data = data.get("credentials", {})
        if data is None:
            raise RuntimeError(f"missing required credentials for organization config with name '{name}'")

        return cls(name, github_id, config_repo, base_template, jsonnet_config, data)

    @classmethod
    def of(
        cls,
        project_name: str,
        github_id: str,
        config_repo: str,
        base_template: str,
        credential_data: dict[str, Any],
        base_dir: str,
        work_dir: str,
    ) -> OrganizationConfig:
        jsonnet_config = JsonnetConfig(
            github_id,
            base_dir,
            base_template,
            False,
            work_dir,
        )

        return cls(project_name, github_id, config_repo, base_template, jsonnet_config, credential_data)


class SecretResolver(Protocol):
    @abstractmethod
    def is_supported_secret_provider(self, provider_type: str) -> bool: ...

    @abstractmethod
    def get_secret(self, data: str) -> str: ...


class CredentialResolver(SecretResolver):
    def __init__(self, config: OtterdogConfig) -> None:
        self._config = config
        self._credential_providers: dict[str, CredentialProvider] = {}

    def _get_credential_provider(self, provider_type: str) -> CredentialProvider | None:
        provider = self._credential_providers.get(provider_type)
        if provider is None:
            provider = CredentialProvider.create(
                provider_type, query_json(f"defaults.{provider_type}", self._config.configuration) or {}
            )
            if provider is not None:
                self._credential_providers[provider_type] = provider

        return provider

    def get_credentials(self, org_config: OrganizationConfig, only_token: bool = False) -> Credentials:
        provider_type = org_config.credential_data.get("provider")

        if provider_type is None:
            provider_type = self._config.default_credential_provider

        if not provider_type:
            raise RuntimeError(f"no credential provider configured for organization '{org_config.name}'")

        provider = self._get_credential_provider(provider_type)
        if provider is not None:
            return provider.get_credentials(org_config.name, org_config.credential_data, only_token)
        else:
            raise RuntimeError(f"unsupported credential provider '{provider_type}'")

    def is_supported_secret_provider(self, provider_type: str) -> bool:
        # TODO: make this cleaner
        return provider_type in ["pass", "bitwarden"]

    def get_secret(self, secret_data: str) -> str:
        if secret_data and ":" in secret_data:
            provider_type, data = re.split(":", secret_data)
            provider = self._get_credential_provider(provider_type)
            if provider is not None:
                return provider.get_secret(data)
            else:
                return secret_data
        else:
            return secret_data


@dataclasses.dataclass(frozen=True)
class OtterdogConfig:
    configuration: Mapping[str, Any]
    local_mode: bool
    working_dir: str

    _base_url: str = dataclasses.field(init=False)
    _jsonnet_config: Mapping[str, Any] = dataclasses.field(init=False)
    _github_config: Mapping[str, Any] = dataclasses.field(init=False)
    _default_credential_provider: str = dataclasses.field(init=False)
    _jsonnet_base_dir: str = dataclasses.field(init=False)

    _organizations_map: dict[str, OrganizationConfig] = dataclasses.field(init=False, default_factory=dict)
    _organizations: list[OrganizationConfig] = dataclasses.field(init=False, default_factory=list)

    def __post_init__(self):
        object.__setattr__(self, "_base_url", query_json("defaults.base_url", self.configuration) or None)
        object.__setattr__(self, "_jsonnet_config", query_json("defaults.jsonnet", self.configuration) or {})
        object.__setattr__(self, "_github_config", query_json("defaults.github", self.configuration) or {})
        object.__setattr__(
            self, "_default_credential_provider", query_json("defaults.credentials.provider", self.configuration) or ""
        )

        object.__setattr__(
            self,
            "_jsonnet_base_dir",
            os.path.join(self.working_dir, self._jsonnet_config.get("config_dir", "orgs")),
        )
        if not os.path.exists(self._jsonnet_base_dir):
            os.makedirs(self._jsonnet_base_dir)

        organizations = self.configuration.get("organizations", [])
        for org in organizations:
            org_config = OrganizationConfig.from_dict(org, self)
            self._organizations.append(org_config)
            self._organizations_map[org_config.name.lower()] = org_config
            self._organizations_map[org_config.github_id.lower()] = org_config

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def jsonnet_base_dir(self) -> str:
        return self._jsonnet_base_dir

    @property
    def default_config_repo(self) -> str:
        return self._github_config.get("config_repo", ".otterdog")

    @property
    def default_team_exclusions(self) -> list[str]:
        return self._github_config.get("exclude_teams", [])

    @cached_property
    def exclude_teams_pattern(self) -> re.Pattern | None:
        team_exclusions = self.default_team_exclusions
        if len(team_exclusions) == 0:
            return None
        else:
            return re.compile("|".join(team_exclusions))

    @property
    def default_credential_provider(self) -> str:
        return self._default_credential_provider

    @property
    def default_base_template(self) -> str:
        base_template = self._jsonnet_config.get("base_template")
        if base_template is None:
            raise RuntimeError(
                "need to define a base template in your otterdog config, key: 'defaults.jsonnet.base_template'"
            )
        return base_template

    def _get_organization_config(self, project_or_organization_name: str):
        return self._organizations_map.get(project_or_organization_name.lower())

    @property
    def project_names(self) -> list[str]:
        return [config.name for config in self._organizations]

    @property
    def organization_names(self) -> list[str]:
        return [config.github_id for config in self._organizations]

    def get_project_name(self, github_id: str) -> str | None:
        org_config = self._get_organization_config(github_id)
        if org_config is not None:
            return org_config.name
        else:
            return None

    def get_organization_config(self, project_or_organization_name: str) -> OrganizationConfig:
        org_config = self._get_organization_config(project_or_organization_name)
        if org_config is None:
            raise RuntimeError(f"unknown organization with name / github_id '{project_or_organization_name}'")
        return org_config

    @classmethod
    def from_file(cls, config_file: str, local_mode: bool, working_dir: str | None = None) -> OtterdogConfig:
        if not os.path.exists(config_file):
            raise RuntimeError(f"configuration file '{config_file}' not found")

        config_file_file = os.path.realpath(config_file)
        config_file_dir = os.path.dirname(config_file)

        with open(config_file_file) as f:
            configuration = json.load(f)

        if working_dir is None:
            override_defaults_file = os.path.join(config_file_dir, ".otterdog-defaults.json")
            if os.path.exists(override_defaults_file):
                with open(override_defaults_file) as defaults_file:
                    defaults = json.load(defaults_file)
                    _logger.trace("loading default overrides from '%s'", override_defaults_file)
                    configuration["defaults"] = deep_merge_dict(defaults, configuration.setdefault("defaults", {}))

        return cls(configuration, local_mode, working_dir if working_dir is not None else config_file_dir)

    @classmethod
    def from_dict(cls, configuration: Mapping[str, Any], local_mode: bool, working_dir: str) -> OtterdogConfig:
        return cls(configuration, local_mode, working_dir)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from . import CacheStrategy

if TYPE_CHECKING:
    from typing import Any

    from aiohttp_client_cache import CacheBackend
    from redis.asyncio.client import Redis


def redis_cache(uri: str, connection: Redis | None = None) -> CacheStrategy:
    return _RedisCache(uri, connection)


class _RedisCache(CacheStrategy):
    def __init__(self, redis_uri: str, connection: Redis | None):
        self._redis_uri = redis_uri
        self._connection = connection

    def get_cache_backend(self) -> CacheBackend:
        from aiohttp_client_cache.backends import RedisBackend

        return RedisBackend(address=self._redis_uri, connection=self._connection)

    def is_external(self) -> bool:
        return False

    def get_request_parameters(self) -> dict[str, Any]:
        return {"refresh": True}

    def __str__(self):
        return f"redis-cache('{self._redis_uri}')"



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class InstallAppOperation(Operation):
    """
    Installs a GitHub app for an organization.
    """

    def __init__(self, app_slug: str):
        super().__init__()
        self._app_slug = app_slug
        self._app_id: int | None = None

    @property
    def app_slug(self) -> str:
        return self._app_slug

    @property
    def app_id(self) -> int | None:
        return self._app_id

    def pre_execute(self) -> None:
        self.printer.println(f"Installing GitHub app '{self.app_slug}':")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id

        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                rest_api = provider.rest_api

                current_app_installations = await rest_api.org.get_app_installations(github_id)
                for app_installation in current_app_installations:
                    installed_app_slug = app_installation["app_slug"]
                    if installed_app_slug == self.app_slug:
                        self.printer.println("app already installed, skipping.")
                        return 0

                org_int_id = str(await rest_api.org.get_id(github_id))
                await provider.web_client.install_github_app(org_int_id, self.app_slug)

            self.printer.println("app installed.")
            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.credentials import CredentialProvider, Credentials
from otterdog.logging import get_logger

if TYPE_CHECKING:
    from typing import Any

_logger = get_logger(__name__)


class PlainVault(CredentialProvider):
    """
    A class to access credentials in clear text.

    NOTE: DO NOT USE THIS PROVIDER unless for quickly testing our otterdog.
    """

    KEY_API_TOKEN = "api_token"
    KEY_USERNAME = "username"
    KEY_PASSWORD = "password"
    KEY_TWOFA_SEED = "twofa_seed"

    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials:
        github_token = self._retrieve_key(self.KEY_API_TOKEN, data)

        if only_token is False:
            username = self._retrieve_key(self.KEY_USERNAME, data)
            password = self._retrieve_key(self.KEY_PASSWORD, data)
            totp_secret = self._retrieve_key(self.KEY_TWOFA_SEED, data)
        else:
            username = None
            password = None
            totp_secret = None

        return Credentials(username, password, totp_secret, github_token)

    def get_secret(self, key_data: str) -> str:
        raise RuntimeError("plain vault does not support secrets")

    @staticmethod
    def _retrieve_key(key: str, data: dict[str, str]) -> str:
        resolved_key = data.get(key)

        if resolved_key is None:
            raise RuntimeError(f"required key '{key}' not found in credential data")
        else:
            return resolved_key

    def __repr__(self):
        return "PlainVault()"



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, cast

from jsonbender import F, Forall, If, K, OptionalS, S  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.providers.github import is_org_settings_key_retrieved_via_web_ui
from otterdog.utils import (
    UNSET,
    Change,
    IndentingPrinter,
    associate_by_key,
    is_set_and_present,
    is_set_and_valid,
    unwrap,
    write_patch_object_as_json,
)

from .custom_property import CustomProperty
from .organization_workflow_settings import OrganizationWorkflowSettings

if TYPE_CHECKING:
    from collections.abc import Iterator

    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class OrganizationSettings(ModelObject):
    """
    Represents settings of a GitHub Organization.
    """

    name: str | None
    plan: str = dataclasses.field(metadata={"read_only": True})
    description: str | None
    email: str | None
    location: str | None
    company: str | None
    billing_email: str
    twitter_username: str | None
    blog: str | None
    has_discussions: bool
    discussion_source_repository: str | None
    has_organization_projects: bool
    default_branch_name: str
    default_repository_permission: str
    two_factor_requirement: bool = dataclasses.field(metadata={"read_only": True})
    web_commit_signoff_required: bool
    default_code_security_configurations_disabled: bool
    members_can_create_private_repositories: bool
    members_can_create_public_repositories: bool
    members_can_fork_private_repositories: bool
    members_can_create_public_pages: bool
    members_can_change_repo_visibility: bool
    members_can_delete_repositories: bool
    members_can_delete_issues: bool
    members_can_create_teams: bool
    readers_can_create_discussions: bool
    packages_containers_public: bool
    packages_containers_internal: bool
    members_can_change_project_visibility: bool
    security_managers: list[str]

    # nested model fields
    workflows: OrganizationWorkflowSettings = dataclasses.field(metadata={"embedded_model": True})
    custom_properties: list[CustomProperty] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)

    @property
    def model_object_name(self) -> str:
        return "settings"

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.has_discussions is False and field.name == "discussion_source_repository":
            return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        # execute custom validation rules if present
        self.execute_custom_validation_if_present(context, "validate-org-settings.py")

        if is_set_and_present(self.description) and len(self.description) > 160:
            context.add_failure(
                FailureType.ERROR,
                "setting 'description' exceeds maximum allowed length of 160 chars.",
            )

        if self.has_discussions is True and not is_set_and_valid(self.discussion_source_repository):
            context.add_failure(
                FailureType.ERROR,
                "enabling 'has_discussions' requires setting a valid repository in 'discussion_source_repository'.",
            )

        if is_set_and_present(self.discussion_source_repository) and "/" not in self.discussion_source_repository:
            context.add_failure(
                FailureType.ERROR,
                "setting 'discussion_source_repository' requires a repository in '<owner>/<repo-name>' format.",
            )

        if is_set_and_valid(self.default_repository_permission):
            if self.default_repository_permission not in {"none", "read", "write", "admin"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'default_repository_permission' has value '{self.default_repository_permission}', "
                    f"while only values ('none' | 'read' | 'write' | 'admin') are allowed.",
                )

        if is_set_and_present(self.custom_properties):
            for custom_property in self.custom_properties:
                custom_property.validate(context, self)

        if is_set_and_present(self.workflows):
            self.workflows.validate(context, self)

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject]]:
        if is_set_and_present(self.custom_properties):
            for custom_property in self.custom_properties:
                yield custom_property, self
                yield from custom_property.get_model_objects()

    @classmethod
    def get_mapping_from_model(cls) -> dict[str, Any]:
        mapping = super().get_mapping_from_model()

        mapping.update(
            {
                "custom_properties": OptionalS("custom_properties", default=[])
                >> Forall(lambda x: CustomProperty.from_model_data(x)),
                "workflows": If(
                    OptionalS("workflows", default=None) == K(None),
                    K(UNSET),
                    S("workflows") >> F(lambda x: OrganizationWorkflowSettings.from_model_data(x)),
                ),
            }
        )

        return mapping

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        mapping["plan"] = OptionalS("plan", "name", default=UNSET)

        if "workflows" in data:
            workflow_data = data["workflows"]
            mapping["workflows"] = K(OrganizationWorkflowSettings.from_provider_data(org_id, workflow_data))
        else:
            mapping["workflows"] = K(UNSET)

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return None

    def changes_require_web_ui(self, changes: dict[str, Change]) -> bool:
        from otterdog.providers.github import is_org_settings_key_retrieved_via_web_ui

        return any(is_org_settings_key_retrieved_via_web_ui(key) for key in changes)

    def unset_settings_requiring_web_ui(self) -> None:
        for key in self.keys():
            if is_org_settings_key_retrieved_via_web_ui(key):
                self.__setattr__(key, UNSET)

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        default_org_settings = cast(OrganizationSettings, default_object)

        patch = self.get_patch_to(default_object)

        if "workflows" in patch:
            patch.pop("workflows")

        write_patch_object_as_json(patch, printer, False)

        # print custom properties
        if is_set_and_present(self.custom_properties) and len(self.custom_properties) > 0:
            properties_by_name = associate_by_key(self.custom_properties, lambda x: x.name)
            default_properties_by_name = associate_by_key(default_org_settings.custom_properties, lambda x: x.name)

            for default_property_name in set(default_properties_by_name):
                if default_property_name in properties_by_name:
                    properties_by_name.pop(default_property_name)

            default_org_custom_property = CustomProperty.from_model_data(config.default_org_custom_property_config)

            if len(properties_by_name) > 0:
                printer.println("custom_properties+: [")
                printer.level_up()

                for custom_property in self.custom_properties:
                    custom_property.to_jsonnet(printer, config, context, False, default_org_custom_property)

                printer.level_down()
                printer.println("],")

        if is_set_and_present(self.workflows):
            default_workflow_settings = cast(OrganizationSettings, default_object).workflows

            patch = self.workflows.get_patch_to(default_workflow_settings)
            if len(patch) > 0:
                printer.print("workflows+:")
                self.workflows.to_jsonnet(printer, config, context, False, default_workflow_settings)

        printer.level_down()
        printer.println("},")

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: OrganizationSettings | None,
        current_object: OrganizationSettings | None,
        parent_object: ModelObject | None,
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        expected_object = unwrap(expected_object)
        current_object = unwrap(current_object)

        modified_settings: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)

        # this setting is only intended to disable any existing default configuration, it can not be enabled per se
        if "default_code_security_configurations_disabled" in modified_settings:
            change: Change[bool] = cast(
                Change[bool], modified_settings.get("default_code_security_configurations_disabled")
            )
            if change.from_value is True and change.to_value is False:
                modified_settings.pop("default_code_security_configurations_disabled")

        if len(modified_settings) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object, current_object, modified_settings, parent_object, False, cls.apply_live_patch
                )
            )

        context.modified_org_settings = modified_settings

        if is_set_and_valid(expected_object.custom_properties):
            CustomProperty.generate_live_patch_of_list(
                expected_object.custom_properties,
                current_object.custom_properties,
                expected_object,
                context,
                handler,
            )

    @classmethod
    async def apply_live_patch(
        cls,
        patch: LivePatch[OrganizationSettings],
        org_id: str,
        provider: GitHubProvider,
    ) -> None:
        if patch.patch_type != LivePatchType.CHANGE:
            raise ValueError(f"unexpected patch_type '{patch.patch_type}'")

        github_settings = await cls.changes_to_provider(org_id, unwrap(patch.changes), provider)

        if "workflows" in github_settings:
            github_settings.pop("workflows")
            update_workflows = True
        else:
            update_workflows = False

        await provider.update_org_settings(
            org_id,
            github_settings,
        )

        if update_workflows is True:
            data = unwrap(patch.expected_object).workflows.to_model_dict(for_diff=True)
            github_data = await OrganizationWorkflowSettings.dict_to_provider_data(org_id, data, provider)

            await provider.update_org_workflow_settings(org_id, github_data)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import logging

import pytest

from otterdog.providers.github.auth import app_auth
from otterdog.providers.github.rest import RestApi

_logger = logging.getLogger(__name__)

_APP_ID = ""
_APP_PRIVATE_KEY = ""


@pytest.mark.asyncio
@pytest.mark.skipif(not _APP_ID, reason="need to fill in app values")
async def test_list_installations():
    async with RestApi(app_auth(_APP_ID, _APP_PRIVATE_KEY)) as rest_api:
        installations = await rest_api.app.get_app_installations()

        for installation in installations:
            permissions = installation["permissions"]
            if "merge_queues" not in permissions:
                _logger.info(f"{installation['account']['login']} -> missing merge_queue permission")



#  *******************************************************************************
#  Copyright (c) 2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from rich import box
from rich.table import Table

from otterdog.utils import unwrap
from otterdog.webapp.db.models import BlueprintStatusModel

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class ListBlueprintsOperation(Operation):
    """
    Lists remediation PRs for blueprints.
    """

    def __init__(self, blueprint_id: str | None):
        super().__init__()
        self._blueprint_id = blueprint_id
        self._table = Table(title="Projects", box=box.ROUNDED)
        self._blueprints_by_org: dict[str, list[BlueprintStatusModel]] = {}

    @property
    def blueprint_id(self) -> str | None:
        return self._blueprint_id

    @property
    def table(self) -> Table:
        return self._table

    def pre_execute(self) -> None:
        import requests

        items = []
        params = {
            "pageIndex": 1,
            "pageSize": 50,
        }

        base_url = unwrap(self._config).base_url
        if base_url is None:
            raise RuntimeError("no base_url set which is required when using operation 'list-blueprints'")

        while True:
            response = requests.get(f"{base_url}/api/blueprints/remediations", params=params, timeout=10)
            response_json = response.json()
            items.extend(response_json["data"])
            if len(items) >= response_json["itemsCount"]:
                break
            else:
                params["pageIndex"] = params["pageIndex"] + 1

        self.table.add_column("Index", justify="left", style="cyan")
        self.table.add_column("Blueprint ID", justify="left", style="cyan", no_wrap=True)
        self.table.add_column("GitHub ID", style="magenta")
        self.table.add_column("Repo", justify="left", style="green")
        self.table.add_column("PR", justify="left", style="red")

        for item in items:
            blueprint_status = BlueprintStatusModel.model_validate(item)

            if self._skip_blueprint(blueprint_status):
                continue

            blueprints = self._blueprints_by_org.setdefault(blueprint_status.id.org_id, [])
            blueprints.append(blueprint_status)

    def _skip_blueprint(self, blueprint: BlueprintStatusModel) -> bool:
        if self.blueprint_id is not None and blueprint.id.blueprint_id != self.blueprint_id:
            return True

        return False

    def post_execute(self) -> None:
        self.printer.console.print(self.table)

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        blueprints = self._blueprints_by_org.get(org_config.github_id, [])
        for blueprint in blueprints:
            pr_url = (
                f"https://github.com/{blueprint.id.org_id}/{blueprint.id.repo_name}/pull/{blueprint.remediation_pr}"
            )

            self.table.add_row(
                f"{len(self.table.rows) + 1}",
                blueprint.id.blueprint_id,
                blueprint.id.org_id,
                blueprint.id.repo_name,
                pr_url,
            )

        return 0



#  *******************************************************************************
#  Copyright (c) 2024-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import re
from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException
from otterdog.providers.github.rest import RestApi, RestClient

_logger = get_logger(__name__)


class TeamClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_team_ids(self, combined_slug: str) -> tuple[int, str]:
        _logger.debug("retrieving team ids for slug '%s'", combined_slug)
        org_id, team_slug = re.split("/", combined_slug)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/teams/{team_slug}")
            return response["id"], response["node_id"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving team node id:\n{ex}") from ex

    async def get_teams(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving teams for org '%s'", org_id)

        try:
            return await self.requester.request_paged_json("GET", f"/orgs/{org_id}/teams")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving teams for org '{org_id}':\n{ex}") from ex

    async def get_team_slugs(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving team slugs for org '%s'", org_id)

        try:
            teams = await self.get_teams(org_id)
            return [team["slug"] for team in teams]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving teams:\n{ex}") from ex

    async def add_team(self, org_id: str, team_name: str, data: dict[str, str]) -> str:
        _logger.debug("adding team '%s' for org '%s'", team_name, org_id)

        status, body = await self.requester.request_raw("POST", f"/orgs/{org_id}/teams", json.dumps(data))

        if status != 201:
            raise RuntimeError(f"failed to add team '{team_name}': {body}")

        team_data = json.loads(body)
        team_slug = team_data["slug"]

        # GitHub automatically adds the creator of the team to the list of maintainers
        # Remove any member of the team right after creation again
        current_members = await self.get_team_members(org_id, team_slug)
        for current_member in current_members:
            await self.remove_member_from_team(org_id, team_slug, current_member["login"])

        if "members" in data:
            members = data["members"]
            for user in members:
                await self.add_member_to_team(org_id, team_slug, user)

        _logger.debug("added team '%s'", team_name)
        return team_slug

    async def update_team(self, org_id: str, team_slug: str, team: dict[str, Any]) -> None:
        _logger.debug("updating team '%s' for org '%s'", team_slug, org_id)

        try:
            await self.requester.request_json("PATCH", f"/orgs/{org_id}/teams/{team_slug}", team)

            if "members" in team:
                await self.update_team_members(org_id, team_slug, team["members"])

            _logger.debug("updated team '%s'", team_slug)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update team '{team_slug}':\n{ex}") from ex

    async def get_team_members(self, org_id: str, team_slug: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving team members for team '%s/%s'", org_id, team_slug)

        try:
            return await self.requester.request_paged_json("GET", f"/orgs/{org_id}/teams/{team_slug}/members")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving team members for team '{org_id}/{team_slug}':\n{ex}") from ex

    async def update_team_members(self, org_id: str, team_slug: str, members: list[str]) -> None:
        _logger.debug("updating team members for team '%s' in org '%s'", team_slug, org_id)

        current_members = {x["login"] for x in await self.get_team_members(org_id, team_slug)}

        # first, add all users that are not members yet.
        for member in members:
            if member in current_members:
                current_members.remove(member)
            else:
                await self.add_member_to_team(org_id, team_slug, member)

        # second, remove the current members that are remaining.
        for member in current_members:
            await self.remove_member_from_team(org_id, team_slug, member)

    async def add_member_to_team(self, org_id: str, team_slug: str, user: str) -> None:
        _logger.debug("adding user with id '%s' to team '%s' in org '%s'", user, team_slug, org_id)

        data = {"role": "member"}
        status, body = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/teams/{team_slug}/memberships/{user}", data=json.dumps(data)
        )

        if status == 200:
            _logger.debug("added user '%s' to team '%s' for org '%s'", user, team_slug, org_id)
        else:
            raise RuntimeError(f"failed adding user '{user}' to team '{team_slug}' in org '{org_id}'\n{status}: {body}")

    async def remove_member_from_team(self, org_id: str, team_slug: str, user: str) -> None:
        _logger.debug("removing user '%s' from team '%s' in org '%s'", user, team_slug, org_id)

        status, body = await self.requester.request_raw(
            "DELETE", f"/orgs/{org_id}/teams/{team_slug}/memberships/{user}"
        )
        if status != 204:
            raise RuntimeError(
                f"failed removing user '{user}' from team '{team_slug}' in org '{org_id}'\n{status}: {body}"
            )

        _logger.debug("removed user '%s' from team '%s' in org '%s'", user, team_slug, org_id)

    async def delete_team(self, org_id: str, team_slug: str) -> None:
        _logger.debug("deleting team '%s' for org '%s'", team_slug, org_id)

        status, body = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/teams/{team_slug}")

        if status != 204:
            raise RuntimeError(f"failed to delete team '{team_slug}': {body}")

        _logger.debug("removed team '%s'", team_slug)

    async def is_user_member_of_team(self, org_id: str, team_slug: str, user: str) -> bool:
        _logger.debug("retrieving membership of user '%s' for team '%s' in org '%s'", user, team_slug, org_id)

        status, body = await self.requester.request_raw("GET", f"/orgs/{org_id}/teams/{team_slug}/memberships/{user}")

        if status == 200:
            return True
        elif status == 404:
            return False
        else:
            raise RuntimeError(
                f"failed retrieving team membership for user '{user}' in org '{org_id}'\n{status}: {body}"
            )

    async def get_membership(self, org_id: str, user_name: str) -> dict[str, Any]:
        _logger.debug("retrieving membership for user '%s' in org '%s'", user_name, org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/memberships/{user_name}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving membership for user '{user_name}' in org '{org_id}':\n{ex}") from ex



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    from aiohttp_client_cache import CacheBackend


class CacheStrategy(ABC):
    @abstractmethod
    def get_cache_backend(self) -> CacheBackend | None: ...

    @abstractmethod
    def is_external(self) -> bool: ...

    def replace_base_url(self, base_url: str) -> str:
        return base_url

    @abstractmethod
    def get_request_parameters(self) -> dict[str, Any]: ...



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
from typing import Any, TypeVar

from otterdog.models import FailureType, ModelObject, ValidationContext

VT = TypeVar("VT", bound="Variable")


@dataclasses.dataclass
class Variable(ModelObject, abc.ABC):
    """
    Represents a Variable.
    """

    name: str = dataclasses.field(metadata={"key": True})
    value: str

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if self.name.startswith("GITHUB_"):
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header(parent_object)} starts with prefix 'GITHUB_', "
                f"which is not allowed for variables.",
            )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
from asyncio import Lock
from functools import cached_property
from shutil import ignore_patterns
from typing import Any

import aiofiles.os
import aiofiles.ospath

from .logging import get_logger
from .utils import jsonnet_evaluate_snippet, parse_github_url, parse_template_url

_template_lock = Lock()
_logger = get_logger(__name__)


class JsonnetConfig:
    # FIXME: the function names to create resources should not be hard-coded but
    #        rather follow a convention to add new resources more easily.

    create_org = "newOrg"
    create_org_role = "newOrgRole"
    create_org_team = "newTeam"
    create_org_custom_property = "newCustomProperty"
    create_org_webhook = "newOrgWebhook"
    create_org_secret = "newOrgSecret"
    create_org_variable = "newOrgVariable"
    create_org_ruleset = "newOrgRuleset"
    create_repo = "newRepo"
    extend_repo = "extendRepo"
    create_repo_webhook = "newRepoWebhook"
    create_repo_secret = "newRepoSecret"
    create_repo_variable = "newRepoVariable"
    create_branch_protection_rule = "newBranchProtectionRule"
    create_repo_ruleset = "newRepoRuleset"
    create_environment = "newEnvironment"
    create_pull_request = "newPullRequest"
    create_status_checks = "newStatusChecks"
    create_merge_queue = "newMergeQueue"

    def __init__(
        self,
        org_id: str,
        base_dir: str,
        base_template_url: str,
        local_only: bool,
        org_dir: str | None = None,
    ):
        self._org_id = org_id
        self._base_dir = base_dir
        self._base_org_dir = org_dir if org_dir is not None else base_dir

        repo_url, file, ref = parse_template_url(base_template_url)

        self._base_template_repo_url = repo_url
        self._base_template_repo_name = os.path.basename(repo_url)
        self._base_template_file = file
        self._base_template_ref = ref

        self._local_only = local_only

        self._default_org_config: dict[str, Any] | None = None
        self._default_org_role_config: dict[str, Any] | None = None
        self._default_org_custom_property_config: dict[str, Any] | None = None
        self._default_org_webhook_config: dict[str, Any] | None = None
        self._default_org_secret_config: dict[str, Any] | None = None
        self._default_org_variable_config: dict[str, Any] | None = None
        self._default_org_ruleset_config: dict[str, Any] | None = None
        self._default_repo_config: dict[str, Any] | None = None
        self._default_repo_webhook_config: dict[str, Any] | None = None
        self._default_repo_secret_config: dict[str, Any] | None = None
        self._default_repo_variable_config: dict[str, Any] | None = None
        self._default_branch_protection_rule_config: dict[str, Any] | None = None
        self._default_repo_ruleset_config: dict[str, Any] | None = None
        self._default_environment_config: dict[str, Any] | None = None
        self._default_pull_request_config: dict[str, Any] | None = None
        self._default_status_checks_config: dict[str, Any] | None = None
        self._default_merge_queue_config: dict[str, Any] | None = None

        self._initialized = False

    @property
    def org_id(self) -> str:
        return self._org_id

    async def init_template(self) -> None:
        if self._initialized is True:
            return

        if not self._local_only:
            await self._init_base_template()

        template_file = self.template_file
        _logger.debug("loading template file '%s'", template_file)
        if not await aiofiles.ospath.exists(self.template_file):
            raise RuntimeError(f"template file '{template_file}' does not exist")

        self._initialized = True

    def default_org_config_for_org_id(self, project_name: str, org_id: str) -> dict[str, Any]:
        try:
            # load the default settings for the organization
            snippet = f"(import '{self.template_file}').{self.create_org}('{project_name}', '{org_id}')"
            return jsonnet_evaluate_snippet(snippet)
        except RuntimeError as ex:
            raise RuntimeError(f"failed to get default organization config for org '{org_id}': {ex}") from ex

    @cached_property
    def default_org_config(self) -> dict[str, Any]:
        return self.default_org_config_for_org_id("default", "default")

    @cached_property
    def default_org_role_config(self):
        try:
            # load the default org role config
            org_role_snippet = f"(import '{self.template_file}').{self.create_org_role}('default')"
            return jsonnet_evaluate_snippet(org_role_snippet)
        except RuntimeError:
            _logger.debug("no default org role config found, roles will be skipped")
            return None

    @cached_property
    def default_team_config(self):
        try:
            # load the default team config
            team_snippet = f"(import '{self.template_file}').{self.create_org_team}('default')"
            return jsonnet_evaluate_snippet(team_snippet)
        except RuntimeError:
            _logger.debug("no default team config found, teams will be skipped")
            return None

    @cached_property
    def default_org_custom_property_config(self):
        try:
            # load the default org custom property config
            org_custom_property_snippet = (
                f"(import '{self.template_file}').{self.create_org_custom_property}('default')"
            )
            return jsonnet_evaluate_snippet(org_custom_property_snippet)
        except RuntimeError:
            _logger.debug("no default org custom property config found, custom properties will be skipped")
            return None

    @cached_property
    def default_org_webhook_config(self):
        try:
            # load the default org webhook config
            org_webhook_snippet = f"(import '{self.template_file}').{self.create_org_webhook}('default')"
            return jsonnet_evaluate_snippet(org_webhook_snippet)
        except RuntimeError:
            _logger.debug("no default org webhook config found, webhooks will be skipped")
            return None

    @cached_property
    def default_org_secret_config(self):
        try:
            # load the default org secret config
            org_secret_snippet = f"(import '{self.template_file}').{self.create_org_secret}('default')"
            return jsonnet_evaluate_snippet(org_secret_snippet)
        except RuntimeError:
            _logger.debug("no default org secret config found, secrets will be skipped")
            return None

    @cached_property
    def default_org_variable_config(self):
        try:
            # load the default org variable config
            org_variable_snippet = f"(import '{self.template_file}').{self.create_org_variable}('default')"
            return jsonnet_evaluate_snippet(org_variable_snippet)
        except RuntimeError:
            _logger.debug("no default org variable config found, variables will be skipped")
            return None

    @cached_property
    def default_org_ruleset_config(self):
        try:
            # load the default org ruleset config
            org_ruleset_snippet = f"(import '{self.template_file}').{self.create_org_ruleset}('default')"
            return jsonnet_evaluate_snippet(org_ruleset_snippet)
        except RuntimeError:
            _logger.debug("no default org ruleset config found, rulesets will be skipped")
            return None

    @cached_property
    def default_repo_config(self):
        try:
            # load the default repo config
            repo_snippet = f"(import '{self.template_file}').{self.create_repo}('default')"
            return jsonnet_evaluate_snippet(repo_snippet)
        except RuntimeError:
            _logger.debug("no default repo config found, repos will be skipped")
            return None

    @cached_property
    def default_repo_webhook_config(self):
        try:
            # load the default repo webhook config
            repo_webhook_snippet = f"(import '{self.template_file}').{self.create_repo_webhook}('default')"
            return jsonnet_evaluate_snippet(repo_webhook_snippet)
        except RuntimeError:
            _logger.debug("no default repo webhook config found, webhooks will be skipped")
            return None

    @cached_property
    def default_repo_secret_config(self):
        try:
            # load the default repo secret config
            repo_secret_snippet = f"(import '{self.template_file}').{self.create_repo_secret}('default')"
            return jsonnet_evaluate_snippet(repo_secret_snippet)
        except RuntimeError:
            _logger.debug("no default repo secret config found, secrets will be skipped")
            return None

    @cached_property
    def default_repo_variable_config(self):
        try:
            # load the default repo variable config
            repo_variable_snippet = f"(import '{self.template_file}').{self.create_repo_variable}('default')"
            return jsonnet_evaluate_snippet(repo_variable_snippet)
        except RuntimeError:
            _logger.debug("no default repo variable config found, variables will be skipped")
            return None

    @cached_property
    def default_branch_protection_rule_config(self):
        try:
            # load the default branch protection rule config
            branch_protection_snippet = (
                f"(import '{self.template_file}').{self.create_branch_protection_rule}('default')"
            )
            return jsonnet_evaluate_snippet(branch_protection_snippet)
        except RuntimeError:
            _logger.debug("no default branch protection rule config found, branch protection rules will be skipped")
            return None

    @cached_property
    def default_repo_ruleset_config(self):
        try:
            # load the default repo ruleset config
            repo_ruleset_snippet = f"(import '{self.template_file}').{self.create_repo_ruleset}('default')"
            return jsonnet_evaluate_snippet(repo_ruleset_snippet)
        except RuntimeError:
            _logger.debug("no default repo ruleset config found, rulesets will be skipped")
            return None

    @cached_property
    def default_environment_config(self):
        try:
            # load the default environment config
            environment_snippet = f"(import '{self.template_file}').{self.create_environment}('default')"
            return jsonnet_evaluate_snippet(environment_snippet)
        except RuntimeError:
            _logger.debug("no default environment config found, environments will be skipped")
            return None

    @cached_property
    def default_pull_request_config(self):
        try:
            # load the default pull request config
            pull_request_snippet = f"(import '{self.template_file}').{self.create_pull_request}()"
            return jsonnet_evaluate_snippet(pull_request_snippet)
        except RuntimeError:
            _logger.debug("no default pull request config found, pull requests will be skipped")
            return None

    @cached_property
    def default_status_checks_config(self):
        try:
            # load the default status check config
            status_checks_snippet = f"(import '{self.template_file}').{self.create_status_checks}()"
            return jsonnet_evaluate_snippet(status_checks_snippet)
        except RuntimeError:
            _logger.debug("no default status checks config found, status checks will be skipped")
            return None

    @cached_property
    def default_merge_queue_config(self):
        try:
            # load the default merge queue config
            merge_queue_snippet = f"(import '{self.template_file}').{self.create_merge_queue}()"
            return jsonnet_evaluate_snippet(merge_queue_snippet)
        except RuntimeError:
            _logger.debug("no default merge queue config found, merge queues will be skipped")
            return None

    @property
    def base_template_repo_name(self) -> str:
        return self._base_template_repo_name

    @property
    def base_template_file_name(self) -> str:
        return self._base_template_file

    @property
    def template_dir(self) -> str:
        return os.path.join(
            self.org_dir,
            "vendor",
            self._base_template_repo_name,
        )

    @property
    def template_file(self) -> str:
        return os.path.join(
            self.template_dir,
            self._base_template_file,
        )

    async def jsonnet_template_files(self):
        import os

        import aiofiles

        for file in await aiofiles.os.listdir(self.template_dir):
            if file.endswith(".libsonnet"):
                yield os.path.join(self.template_dir, file)

    @property
    def base_dir(self) -> str:
        return self._base_dir

    @property
    def base_org_dir(self) -> str:
        return self._base_org_dir

    @property
    def org_dir(self) -> str:
        return f"{self.base_org_dir}/{self.org_id}"

    @property
    def org_config_file(self) -> str:
        return f"{self.org_dir}/{self.org_id}.jsonnet"

    @property
    def import_statement(self) -> str:
        return f"import 'vendor/{self._base_template_repo_name}/{self._base_template_file}'"

    async def _init_base_template(self) -> None:
        import git
        from aiofiles.os import makedirs
        from aiofiles.ospath import exists
        from aioshutil import copytree, rmtree

        _logger.debug("initializing base template '%s@%s'", self._base_template_repo_url, self._base_template_ref)
        template_owner, template_repository = parse_github_url(self._base_template_repo_url)

        async with _template_lock:
            # cache the template repo with the requested ref in the 'templates' directory
            template_dir = f"{self.base_dir}/templates/{template_owner}/{template_repository}/{self._base_template_ref}"

            if not await exists(f"{template_dir}/.git"):
                _logger.debug("cloning base template from url '%s'", self._base_template_repo_url)
                repo = git.Repo.clone_from(self._base_template_repo_url, template_dir)
                repo.git.checkout(self._base_template_ref)
            else:
                repo = git.Repo(template_dir)
                if not repo.head.is_detached:
                    _logger.debug(
                        "pulling changes from base template url '%s' with ref '%s'",
                        self._base_template_repo_url,
                        repo.head.ref,
                    )
                    repo.remotes.origin.pull()

        # create base directory if it does not exist yet
        if not await exists(self.org_dir):
            await makedirs(self.org_dir)

        if await exists(f"{self.org_dir}/vendor"):
            await rmtree(f"{self.org_dir}/vendor")

        # copy over the cloned template repository
        base_template_directory = os.path.dirname(self.base_template_file_name)
        if len(base_template_directory) > 0:
            # if the base template is in a subdir, only copy this subdir
            pattern = f"{base_template_directory}/**"

            def _include_pattern(path, names):
                from pathlib import PurePath

                ignored_names = []
                for name in names:
                    p = PurePath(os.path.join(path, name))
                    relative_path = p.relative_to(template_dir)
                    if not base_template_directory.startswith(str(relative_path)) and not p.match(pattern):
                        ignored_names.append(name)
                return set(ignored_names)

            ignore = _include_pattern
        else:
            ignore = ignore_patterns(".git")
        await copytree(template_dir, self.template_dir, ignore=ignore)

    def __repr__(self) -> str:
        return f"JsonnetConfig('{self.base_dir}, '{self._base_template_file}')"



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import logging
from typing import cast

from rich.box import Box
from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

TRACE = logging.DEBUG - 5

# verbose levels
# 0: warning
# 1: info
# 2: debug
# 3: trace
_verbose_level = 0

CONSOLE_STDOUT = Console(
    theme=Theme(
        {
            "logging.level.error": "red",
            "logging.level.warning": "yellow",
            "logging.level.info": "green",
            "logging.level.debug": "cyan",
            "logging.level.trace": "magenta",
        }
    ),
    highlight=False,
)
CONSOLE_STDERR = Console(stderr=True)


class CustomLogger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        logging.addLevelName(TRACE, "TRACE")

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)


logging.setLoggerClass(CustomLogger)


def init_logging(verbose: int, setup_python_logger: bool = True) -> None:
    global _verbose_level
    _verbose_level = verbose

    if verbose < 0:
        raise RuntimeError(f"negative verbose level not valid: {verbose}")

    show_time = False

    match verbose:
        case 0:
            level = logging.WARNING
        case 1:
            level = logging.INFO
        case 2:
            level = logging.DEBUG
        case 3:
            level = TRACE
        case _:
            level = TRACE
            show_time = True

    if setup_python_logger is True:
        logging.basicConfig(
            level=logging.WARNING,
            format="%(message)s",
            datefmt="%X.%f",
            handlers=[
                RichHandler(
                    rich_tracebacks=True,
                    tracebacks_show_locals=True,
                    show_time=show_time,
                    omit_repeated_times=False,
                    console=CONSOLE_STDOUT,
                )
            ],
        )

    import otterdog

    logging.getLogger(otterdog.__name__).setLevel(level)


def get_logger(name: str) -> CustomLogger:
    return cast(CustomLogger, logging.getLogger(name))


def is_info_enabled() -> bool:
    return _verbose_level >= 1


def is_debug_enabled() -> bool:
    return _verbose_level >= 2


def is_trace_enabled() -> bool:
    return _verbose_level >= 3


def print_exception(exc: Exception) -> None:
    if is_debug_enabled():
        import asyncio

        from rich.traceback import Traceback

        def get_rich_tb(exception: Exception) -> Traceback:
            return Traceback.from_exception(
                type(exception),
                exception,
                exception.__traceback__,
                show_locals=is_trace_enabled(),
                suppress=[asyncio],
                width=None,
            )

        if isinstance(exc, ExceptionGroup):
            CONSOLE_STDERR.print(get_rich_tb(exc))
            for nested_exception in exc.exceptions:
                CONSOLE_STDERR.print(get_rich_tb(nested_exception))
        else:
            CONSOLE_STDERR.print(get_rich_tb(exc))
    else:
        if isinstance(exc, ExceptionGroup):
            for nested_exception in exc.exceptions:
                print_error(str(nested_exception))
        else:
            print_error(str(exc))


def print_info(msg: str, console: Console = CONSOLE_STDOUT) -> None:
    if is_info_enabled():
        _print_message(msg, "green", "Info", console)


def print_warn(msg: str, console: Console = CONSOLE_STDOUT) -> None:
    _print_message(msg, "yellow", "Warning", console)


def print_error(msg: str, console: Console = CONSOLE_STDOUT) -> None:
    _print_message(msg, "red", "Error", console)


# fmt: off
_DEFAULT_BOX: Box = Box(
    "╷   \n"
    "│   \n"
    "│   \n"
    "│   \n"
    "│   \n"
    "│   \n"
    "│   \n"
    "╵   \n",
    ascii=False,
)
# fmt: on


def _print_message(msg: str, color: str, level: str, console: Console, custom_prefix: str | None = None) -> None:
    from rich.table import Table

    if custom_prefix is not None:
        # fmt: off
        custom_box: Box = Box(
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n"
            f"{custom_prefix}   \n",
            ascii=True,
        )
        box = custom_box
    else:
        box = _DEFAULT_BOX

    table = Table(show_header=False, show_footer=False, show_lines=False, box=box, border_style=color)
    table.add_column("severity", justify="left", style=color)
    table.add_column("message", justify="left", no_wrap=False)
    table.add_row(level + ":", msg)
    console.print(table)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import json
from functools import cache
from typing import TYPE_CHECKING

from aiohttp.client import ClientSession, ClientTimeout, TCPConnector
from aiohttp_retry import ExponentialRetry, RetryClient

from otterdog.logging import get_logger, is_trace_enabled
from otterdog.providers.github.stats import RequestStatistics
from otterdog.utils import query_json

if TYPE_CHECKING:
    from typing import Any

    from otterdog.providers.github.auth import AuthStrategy
    from otterdog.providers.github.cache import CacheStrategy

_logger = get_logger(__name__)


class GraphQLClient:
    _GH_GRAPHQL_URL_ROOT = "api.github.com/graphql"

    def __init__(self, auth_strategy: AuthStrategy, cache_strategy: CacheStrategy | None = None):
        self._auth = auth_strategy.get_auth()

        self._headers = {
            "X-Github-Next-Global-ID": "1",
        }

        self._statistics = RequestStatistics()

        self._cache_strategy = cache_strategy

        if cache_strategy is not None and cache_strategy.is_external():
            self._base_url = cache_strategy.replace_base_url(f"https://{self._GH_GRAPHQL_URL_ROOT}")
            self._use_proxy = True
        else:
            self._base_url = f"https://{self._GH_GRAPHQL_URL_ROOT}"
            self._use_proxy = False

        self._session = ClientSession(
            timeout=ClientTimeout(connect=3, sock_connect=3),
            connector=TCPConnector(limit=10),
        )

        self._client = RetryClient(
            retry_options=ExponentialRetry(3, exceptions={Exception}),
            client_session=self._session,
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exception_type, exception_value, exception_traceback):
        await self.close()

    async def close(self) -> None:
        await self._session.close()

    @property
    def statistics(self) -> RequestStatistics:
        return self._statistics

    async def get_branch_protection_rule_id(self, org_id: str, repo_name: str, pattern: str) -> str:
        _logger.debug(f"getting branch protection rule id for pattern '{pattern}' at repo '{org_id}/{repo_name}'")

        variables = {"organization": org_id, "repository": repo_name}
        branch_protection_rules = await self._run_paged_query(variables, "get-branch-protection-rule-ids.gql")

        for rule in branch_protection_rules:
            if rule["pattern"] == pattern:
                return rule["id"]

        raise RuntimeError(f"failed to find branch protection rule with pattern '{pattern}'")

    async def get_branch_protection_rules(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        _logger.debug(f"retrieving branch protection rules for repo '{org_id}/{repo_name}'")

        variables = {"organization": org_id, "repository": repo_name}
        branch_protection_rules = await self._run_paged_query(variables, "get-branch-protection-rules.gql")

        for branch_protection_rule in branch_protection_rules:
            await self._fill_paged_results_if_needed(
                branch_protection_rule,
                "pushAllowances",
                "pushRestrictions",
                "get-push-allowances.gql",
            )

            await self._fill_paged_results_if_needed(
                branch_protection_rule,
                "reviewDismissalAllowances",
                "reviewDismissalAllowances",
                "get-review-dismissal-allowances.gql",
            )

            await self._fill_paged_results_if_needed(
                branch_protection_rule,
                "bypassPullRequestAllowances",
                "bypassPullRequestAllowances",
                "get-bypass-pull-request-allowances.gql",
            )

            await self._fill_paged_results_if_needed(
                branch_protection_rule,
                "bypassForcePushAllowances",
                "bypassForcePushAllowances",
                "get-bypass-force-push-allowances.gql",
            )

        return branch_protection_rules

    async def _fill_paged_results_if_needed(
        self,
        branch_protection_rule: dict[str, Any],
        input_key: str,
        output_key: str,
        query_file: str,
    ):
        value = branch_protection_rule.pop(input_key)

        nodes = value["nodes"]
        all_actors = self._transform_actors(nodes)
        has_more = bool(query_json("pageInfo.hasNextPage", value) or False)

        if has_more is True:
            variables = {"branchProtectionRuleId": branch_protection_rule["id"]}
            more_actors = await self._run_paged_query(variables, query_file, f"data.node.{input_key}")
            all_actors.extend(self._transform_actors(more_actors))

        branch_protection_rule[output_key] = all_actors

    async def update_branch_protection_rule(
        self,
        org_id: str,
        repo_name: str,
        rule_pattern: str,
        rule_id: str,
        data: dict[str, Any],
    ) -> None:
        _logger.debug(f"updating branch protection rule '{rule_pattern}' for repo '{org_id}/{repo_name}'")

        data["branchProtectionRuleId"] = rule_id
        variables = {"ruleInput": data}

        query = """mutation($ruleInput: UpdateBranchProtectionRuleInput!) {
           updateBranchProtectionRule(input:$ruleInput) {
             branchProtectionRule {
               pattern
             }
           }
        }"""

        status, body = await self._request_raw("POST", query=query, variables=variables)
        if status >= 400:
            raise RuntimeError(
                f"failed updating branch protection rule '{rule_pattern}' for repo '{repo_name}': {body}"
            )

        json_data = json.loads(body)
        if "data" not in json_data:
            raise RuntimeError(f"failed to update branch protection rule '{rule_pattern}': {body}")

        _logger.debug(f"successfully updated branch protection rule '{rule_pattern}'")

    async def add_branch_protection_rule(
        self, org_id: str, repo_name: str, repo_node_id: str, data: dict[str, Any]
    ) -> None:
        rule_pattern = data["pattern"]
        _logger.debug(
            f"creating branch_protection_rule with pattern '{rule_pattern}' " f"for repo '{org_id}/{repo_name}'"
        )

        data["repositoryId"] = repo_node_id
        variables = {"ruleInput": data}

        query = """mutation($ruleInput: CreateBranchProtectionRuleInput!) {
           createBranchProtectionRule(input:$ruleInput) {
             branchProtectionRule {
               pattern
               blocksCreations
             }
           }
        }"""

        status, body = await self._request_raw("POST", query, variables)
        if status >= 400:
            raise RuntimeError(
                f"failed creating branch protection rule '{rule_pattern}' for repo '{repo_name}': {body}"
            )

        json_data = json.loads(body)
        if "data" not in json_data:
            raise RuntimeError(f"failed to create branch protection rule '{rule_pattern}': {body}")

        _logger.debug(f"successfully created branch protection rule '{rule_pattern}'")

    async def delete_branch_protection_rule(self, org_id: str, repo_name: str, rule_pattern: str, rule_id: str) -> None:
        _logger.debug(f"deleting branch protection rule '{rule_pattern}' for repo '{org_id}/{repo_name}'")

        variables = {"ruleInput": {"branchProtectionRuleId": rule_id}}

        query = """mutation($ruleInput: DeleteBranchProtectionRuleInput!) {
           deleteBranchProtectionRule(input:$ruleInput) {
             clientMutationId
           }
        }"""

        status, body = await self._request_raw("POST", query, variables)
        if status >= 400:
            raise RuntimeError(
                f"failed removing branch protection rule '{rule_pattern}' for repo '{repo_name}': {body}"
            )

        _logger.debug(f"successfully removed branch protection rule '{rule_pattern}'")

    async def get_issue_comments(self, org_id: str, repo_name: str, issue_number: int) -> list[dict[str, Any]]:
        _logger.debug(f"retrieving issue comments for issue '{org_id}/{repo_name}/#{issue_number}'")

        variables = {"owner": org_id, "repo": repo_name, "number": issue_number}
        issue_comments = await self._run_paged_query(
            variables, "get-issue-comments.gql", "data.repository.pullRequest.comments"
        )
        return issue_comments

    async def minimize_comment(self, comment_id: str, classifier: str) -> None:
        _logger.debug("minimizing comment")

        variables = {"input": {"subjectId": comment_id, "classifier": classifier}}

        query = """mutation($input: MinimizeCommentInput!) {
           minimizeComment(input: $input) {
             clientMutationId
           }
        }"""

        status, body = await self._request_raw("POST", query, variables)
        if status >= 400:
            raise RuntimeError(f"failed minimizing comment: {body}")

        _logger.debug("successfully minimized comment")

    async def get_team_membership(self, org_id: str, user_login: str) -> list[dict[str, Any]]:
        _logger.debug(f"retrieving team membership for user '{user_login}' in org '{org_id}'")

        variables = {"owner": org_id, "user": user_login}
        return await self._run_paged_query(variables, "get-team-membership.gql", "data.organization.teams")

    async def _run_paged_query(
        self,
        input_variables: dict[str, Any],
        query_file: str,
        prefix_selector: str = "data.repository.branchProtectionRules",
    ) -> list[dict[str, Any]]:
        _logger.debug(f"running graphql query '{query_file}' with input '{json.dumps(input_variables)}'")

        query = _get_query_from_file(query_file)

        finished = False
        end_cursor = None
        result = []

        while not finished:
            variables = {"endCursor": end_cursor}
            variables.update(input_variables)

            status, body = await self._request_raw("POST", query, variables)
            json_data = json.loads(body)

            if is_trace_enabled():
                _logger.trace("graphql result = %s", json.dumps(json_data, indent=2))

            if "data" in json_data:
                rules_result = query_json(prefix_selector + ".nodes", json_data)

                for rule in rules_result:
                    result.append(rule)

                page_info = query_json(prefix_selector + ".pageInfo", json_data)

                if page_info["hasNextPage"]:
                    end_cursor = page_info["endCursor"]
                else:
                    finished = True
            else:
                raise RuntimeError(f"failed running graphql query '{query_file}': {body}")

        return result

    async def _request_raw(self, method: str, query: str, variables: dict[str, Any]) -> tuple[int, str]:
        _logger.trace("'%s', query = %s, variables = %s", method, query[0:300] + "...", variables)

        headers = self._headers.copy()
        if self._auth is not None:
            self._auth.update_headers_with_authorization(headers)

        if self._cache_strategy is not None and self._use_proxy is True:
            kwargs = self._cache_strategy.get_request_parameters()
        else:
            kwargs = {}

        async with self._client.request(
            method,
            url=self._base_url,
            headers=headers,
            json={"query": query, "variables": variables},
            **kwargs,
        ) as response:
            self._statistics.sent_request()

            text = await response.text()
            status = response.status

            if status == 403 or status == 429:
                _logger.trace(f"graphql '{method}' result = ({status}, {text})")
                raise RuntimeError("failed running graphql query, hitting rate limit")

            self._statistics.update_remaining_rate_limit(int(response.headers.get("x-ratelimit-remaining", -1)))
            return status, text

    @staticmethod
    def _transform_actors(actors: list[dict[str, Any]]) -> list[str]:
        result = []
        for actor_wrapper in actors:
            actor = actor_wrapper["actor"]
            if actor is None:
                continue

            typename = actor["__typename"]
            if typename == "User":
                login = actor["login"]
                result.append(f"@{login}")
            elif typename == "Team":
                slug = actor["combinedSlug"]
                result.append(f"@{slug}")
            elif typename == "App":
                slug = actor["slug"]
                result.append(slug)
            else:
                raise RuntimeError(f"unsupported actor '{actor}'")

        return result


@cache
def _get_query_from_file(query_file: str) -> str:
    from importlib_resources import files

    from otterdog import resources

    return files(resources).joinpath(f"graphql/{query_file}").read_text()



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import filecmp
from typing import TYPE_CHECKING

from aiofiles import open
from git import InvalidGitRepositoryError, Repo
from git.config import GitConfigParser

from otterdog.providers.github import GitHubProvider
from otterdog.utils import get_approval

from . import Operation
from .local_plan import LocalPlanOperation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class PushOperation(Operation):
    """
    Pushes a local configuration of an organization to its meta-data repository.
    """

    def __init__(self, show_diff: bool, force_processing: bool, push_message: str | None):
        super().__init__()
        self._show_diff = show_diff
        self._force_processing = force_processing
        self._push_message = push_message

    @property
    def show_diff(self) -> bool:
        return self._show_diff

    @property
    def push_message(self) -> str | None:
        return self._push_message

    @property
    def force_processing(self) -> bool:
        return self._force_processing

    def pre_execute(self) -> None:
        self.printer.println("Pushing organization configurations:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()

        self._print_project_header(org_config, org_index, org_count)

        org_file_name = jsonnet_config.org_config_file
        if not await self.check_config_file_exists(org_file_name):
            return 1

        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            # try to determine the git user configuration if possible
            # if no configuration can be found, omit adding author information

            try:
                git_config_reader = Repo(self.config.working_dir).config_reader()
            except InvalidGitRepositoryError:
                # if the config dir is not a git repo, just read the global config
                git_config_reader = GitConfigParser(None, read_only=True)

            if git_config_reader.has_option("user", "name"):
                author_name = git_config_reader.get_value("user", "name")
            else:
                author_name = None

            if git_config_reader.has_option("user", "email"):
                author_email = git_config_reader.get_value("user", "email")
            else:
                author_email = None

            if not author_name:
                author_name = None
                author_email = None

            async with open(org_file_name) as file:
                content = await file.read()

            async with GitHubProvider(credentials) as provider:
                try:
                    perform_update = True
                    updated_files = []

                    if self.show_diff:
                        perform_update = await self._display_diff(org_config, provider)

                    if perform_update:
                        if await provider.update_content(
                            org_config.github_id,
                            org_config.config_repo,
                            f"otterdog/{github_id}.jsonnet",
                            content,
                            None,
                            self.push_message,
                            author_name,
                            author_email,
                        ):
                            updated_files.append(f"otterdog/{github_id}.jsonnet")

                except RuntimeError as e:
                    self.printer.print_error(
                        f"failed to push definition to repo '{org_config.github_id}/{org_config.config_repo}': {e!s}"
                    )
                    return 1

            if len(updated_files) > 0:
                self.printer.println(
                    f"organization definition pushed to repo '{org_config.github_id}/{org_config.config_repo}': "
                )
                for updated_file in updated_files:
                    self.printer.println(f"  - '{updated_file}'")
            else:
                self.printer.println("no changes, nothing pushed")

            return 0
        finally:
            self.printer.level_down()

    async def _display_diff(self, org_config: OrganizationConfig, provider: GitHubProvider) -> bool:
        rest_api = provider.rest_api
        github_id = org_config.github_id

        default_branch = await rest_api.repo.get_default_branch(org_config.github_id, org_config.config_repo)

        try:
            current_definition = await provider.get_content(
                github_id,
                org_config.config_repo,
                f"otterdog/{github_id}.jsonnet",
                default_branch,
            )
        except RuntimeError:
            if not self.force_processing:
                self.printer.println()
                self.printer.println("No configuration yet available.")
                self.printer.println(
                    "Do you want to push these changes? " "(Only 'yes' or 'y' will be accepted as approval)\n"
                )

                self.printer.print("[bold]Enter a value:[/] ")
                if not get_approval():
                    self.printer.println("\npush cancelled.")
                    return False

            return True

        current_config_file = org_config.jsonnet_config.org_config_file + "-BASE"
        async with open(current_config_file, "w") as file:
            await file.write(current_definition)

        if filecmp.cmp(current_config_file, org_config.jsonnet_config.org_config_file):
            return False

        self.printer.println("The following changes compared to the current configuration exist locally:")
        self.printer.println()
        self.printer.level_up()

        try:
            operation = LocalPlanOperation("-BASE", "*", False, False, "")
            operation.init(self.config, self.printer)
            valid_config = await operation.generate_diff(org_config)
        finally:
            self.printer.level_down()

        if valid_config != 0:
            return False

        if not self.force_processing:
            self.printer.println()
            self.printer.println(
                "Do you want to push these changes? " "(Only 'yes' or 'y' will be accepted as approval)\n"
            )

            self.printer.print("[bold]Enter a value:[/] ")
            if not get_approval():
                self.printer.println("\npush cancelled.")
                return False

        return True



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from otterdog.models import LivePatch, LivePatchType
from otterdog.models.webhook import Webhook
from otterdog.utils import unwrap

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class OrganizationWebhook(Webhook):
    """
    Represents a Webhook defined on organization level.
    """

    @property
    def model_object_name(self) -> str:
        return "org_webhook"

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_org_webhook}"

    @classmethod
    async def apply_live_patch(
        cls,
        patch: LivePatch[OrganizationWebhook],
        org_id: str,
        provider: GitHubProvider,
    ) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                await provider.add_org_webhook(
                    org_id,
                    await unwrap(patch.expected_object).to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                current_object = unwrap(patch.current_object)
                await provider.delete_org_webhook(
                    org_id,
                    current_object.id,
                    current_object.url,
                )

            case LivePatchType.CHANGE:
                current_object = unwrap(patch.current_object)
                await provider.update_org_webhook(
                    org_id,
                    current_object.id,
                    current_object.url,
                    await unwrap(patch.expected_object).to_provider_data(org_id, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import unittest
from io import StringIO
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from otterdog.operations import Operation
from otterdog.utils import Change, IndentingPrinter, LogLevel


class OperationTest(unittest.TestCase):
    # fmt: off
    @parameterized.expand(
        [
            (
                {"A": Change("1", "2"),
                 "B": Change("Hello", "World")},
                '~  {'
                '  ~ A = "1" -> "2"'
                '  ~ B = "Hello" -> "World"'
                '~ }'),
            (
                {"A": Change([1, 2], [3]),
                 "B": Change(3, 4)},
                '~  {'
                '  ~ A = ['
                '    ~ 1 -> 3'
                '    - 2'
                '  ~ ]'
                '  ~ B = 3 -> 4'
                '~ }',
            ),
            (
                {"A": Change({"B": [1, 2]}, {"B": [3]})},
                '~  {'
                '  ~ A   = {'
                '    ~ B   = ['
                '      ~ 1 -> 3'
                '      - 2'
                '    ~ ]'
                '  ~ }'
                '~ }',
            ),
        ]
    )
    # fmt: on
    @patch.multiple(Operation, __abstractmethods__=set())
    def test_print_modified_dict(self, test_input, expected: str):
        operation = Operation()  # type: ignore

        output = StringIO()
        printer = IndentingPrinter(output, log_level=LogLevel.ERROR)
        operation.init(None, printer)  # type: ignore

        operation.print_modified_dict(test_input, "", False)

        assert _strip_string(output.getvalue()) == expected


def _strip_string(input_string: str) -> str:
    return _remove_lf(_remove_ansi_color_codes(input_string))


def _remove_ansi_color_codes(input_string: str) -> str:
    import re

    # 7-bit and 8-bit C1 ANSI sequences
    ansi_escape_8bit = re.compile(r"\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~]")
    return ansi_escape_8bit.sub("", input_string)


def _remove_lf(input_string: str) -> str:
    return input_string.replace("\n", "")



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, cast

from jsonbender import S  # type: ignore

from otterdog.models import (
    FailureType,
    ModelObject,
    ValidationContext,
)
from otterdog.models.workflow_settings import WorkflowSettings
from otterdog.utils import UNSET, is_set_and_valid, is_unset

if TYPE_CHECKING:
    from otterdog.models.organization_workflow_settings import OrganizationWorkflowSettings
    from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class RepositoryWorkflowSettings(WorkflowSettings):
    """
    Represents workflow settings defined on repository level.
    """

    enabled: bool

    def coerce_from_org_settings(
        self, parent_object: ModelObject | None, org_workflow_settings: OrganizationWorkflowSettings
    ) -> RepositoryWorkflowSettings:
        copy = dataclasses.replace(self)

        if org_workflow_settings.enabled_repositories == "none":
            copy.enabled = UNSET  # type: ignore

        from otterdog.models.repository import Repository

        repository_name = cast(Repository, parent_object).name

        if (
            org_workflow_settings.enabled_repositories == "selected"
            and repository_name not in org_workflow_settings.selected_repositories
        ):
            copy.enabled = UNSET  # type: ignore

        if org_workflow_settings.are_actions_more_or_equally_restricted(copy.allowed_actions):
            copy.allowed_actions = UNSET  # type: ignore
            for prop in self._selected_action_properties:
                copy.__setattr__(prop, UNSET)

        if org_workflow_settings.default_workflow_permissions == "read":
            copy.default_workflow_permissions = UNSET  # type: ignore

        if org_workflow_settings.actions_can_approve_pull_request_reviews is False:
            copy.actions_can_approve_pull_request_reviews = UNSET  # type: ignore

        return copy

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        from .repository import Repository

        super().validate(context, parent_object)

        repo = cast(Repository, parent_object)

        actions_enabled = None
        if is_set_and_valid(self.enabled) and self.enabled is True:
            from .github_organization import GitHubOrganization

            actions_enabled = True
            org_workflow_settings = cast(GitHubOrganization, context.root_object).settings.workflows

            if org_workflow_settings.enabled_repositories == "none":
                actions_enabled = False
                context.add_failure(
                    FailureType.INFO,
                    f"{parent_object.get_model_header()} has enabled workflows, "
                    f"while on organization level it is disabled for all repositories, setting will be ignored.",
                )

            if (
                org_workflow_settings.enabled_repositories == "selected"
                and repo.name not in org_workflow_settings.selected_repositories
            ):
                actions_enabled = False
                context.add_failure(
                    FailureType.INFO,
                    f"{parent_object.get_model_header()} has enabled workflows, "
                    f"while on organization level it is only enabled for selected repositories, "
                    f"setting will be ignored.",
                )

            if org_workflow_settings.are_actions_more_restricted(self.allowed_actions):
                context.add_failure(
                    FailureType.INFO,
                    f"{parent_object.get_model_header()} has set 'allowed_actions' to '{self.allowed_actions}', "
                    f"while on organization level it is more restricted to '{org_workflow_settings.allowed_actions}', "
                    f"setting will be ignored.",
                )

            if (
                org_workflow_settings.default_workflow_permissions == "read"
                and self.default_workflow_permissions == "write"
            ):
                context.add_failure(
                    FailureType.INFO,
                    f"{parent_object.get_model_header()} has 'default_workflow_permissions' of value "
                    f"'{self.default_workflow_permissions}', "
                    f"while on organization level it is restricted to "
                    f"'{org_workflow_settings.default_workflow_permissions}', setting will be ignored.",
                )

            if (
                org_workflow_settings.actions_can_approve_pull_request_reviews is False
                and self.actions_can_approve_pull_request_reviews is True
            ):
                context.add_failure(
                    FailureType.INFO,
                    f"{parent_object.get_model_header()} has 'actions_can_approve_pull_request_reviews' enabled, "
                    f"while on organization level it is disabled, setting will be ignored.",
                )
        elif self.enabled is False:
            actions_enabled = False

        if repo.code_scanning_default_setup_enabled is True and actions_enabled is False:
            context.add_failure(
                FailureType.ERROR,
                f"{parent_object.get_model_header()} has 'code_scanning_default_setup_enabled' of "
                f"value '{repo.code_scanning_default_setup_enabled}' while GitHub Actions are disabled.",
            )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if is_unset(self.enabled):
            return False

        if self.enabled is False:
            if field.name == "enabled":
                return True
            else:
                return False

        return super().include_field_for_diff_computation(field)

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        if "enabled" in data and data["enabled"] is False:
            return {"enabled": S("enabled")}
        else:
            return await super().get_mapping_to_provider(org_id, data, provider)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol

from otterdog.logging import get_logger, print_warn

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from typing import Any

_logger = get_logger(__name__)


@dataclasses.dataclass
class Credentials:
    """
    A simple data class to hold credential information to access GitHub.
    """

    _username: str | None
    _password: str | None
    _totp_secret: str | None
    _github_token: str | None

    _last_totp: str | None = None

    @property
    def username(self) -> str:
        if self._username is None:
            raise RuntimeError("username not available")
        else:
            return self._username

    @property
    def password(self) -> str:
        if self._password is None:
            raise RuntimeError("password not available")
        else:
            return self._password

    @property
    def totp(self) -> str:
        import time

        import mintotp  # type: ignore

        if self._totp_secret is None:
            raise RuntimeError("totp_secret not available")

        while True:
            totp = mintotp.totp(self._totp_secret)
            _logger.trace("generated totp '%s'", totp)

            if self._last_totp is None or totp != self._last_totp:
                self._last_totp = totp
                return totp
            else:
                _logger.info("waiting 3s till generating new totp ...")
                time.sleep(3)

    @property
    def github_token(self) -> str:
        if self._github_token is None:
            raise RuntimeError("github_token not available")
        else:
            return self._github_token

    def __str__(self) -> str:
        return f"Credentials(username={self.username})"


class CredentialProvider(Protocol):
    @abstractmethod
    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials: ...

    @abstractmethod
    def get_secret(self, data: str) -> str: ...

    @classmethod
    def create(cls, provider_type: str, defaults: Mapping[str, Any]) -> CredentialProvider | None:
        match provider_type:
            case "bitwarden":
                from .bitwarden_provider import BitwardenVault

                valid_keys = _check_valid_keys(provider_type, defaults, BitwardenVault.__init__)
                return BitwardenVault(**valid_keys)

            case "pass":
                from .pass_provider import PassVault

                valid_keys = _check_valid_keys(provider_type, defaults, PassVault.__init__)
                return PassVault(**valid_keys)

            case "inmemory":
                from .inmemory_provider import InMemoryVault

                _check_valid_keys(provider_type, defaults, InMemoryVault.__init__)
                return InMemoryVault()

            case "plain":
                from .plain_provider import PlainVault

                _check_valid_keys(provider_type, defaults, PlainVault.__init__)
                return PlainVault()

            case _:
                return None


def _check_valid_keys(provider_type: str, defaults: Mapping[str, Any], func: Callable) -> dict[str, Any]:
    import inspect

    result = {}
    signature = inspect.signature(func)
    for k, v in defaults.items():
        if k in signature.parameters:
            result[k] = v
        else:
            print_warn(f"found unexpected key/value pair '{k}:{v}' in defaults for provider '{provider_type}'")

    return result



#  *******************************************************************************
#  Copyright (c) 2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import unwrap
from otterdog.webapp.db.models import BlueprintStatusModel

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class ApproveBlueprintsOperation(Operation):
    """
    Approves and merges remediation PRs for blueprints.
    """

    def __init__(self, blueprint_id: str | None):
        super().__init__()
        self._blueprint_id = blueprint_id
        self._blueprints_by_org: dict[str, list[BlueprintStatusModel]] = {}

    @property
    def blueprint_id(self) -> str | None:
        return self._blueprint_id

    def pre_execute(self) -> None:
        import requests

        items = []
        params = {
            "pageIndex": 1,
            "pageSize": 50,
        }

        base_url = unwrap(self._config).base_url
        if base_url is None:
            raise RuntimeError("no base_url set which is required when using operation 'approve-blueprints'")

        while True:
            response = requests.get(f"{base_url}/api/blueprints/remediations", params=params, timeout=10)
            response_json = response.json()
            items.extend(response_json["data"])
            if len(items) >= response_json["itemsCount"]:
                break
            else:
                params["pageIndex"] = params["pageIndex"] + 1

        for item in items:
            blueprint_status = BlueprintStatusModel.model_validate(item)

            if self._skip_blueprint(blueprint_status):
                continue

            blueprint_list = self._blueprints_by_org.setdefault(blueprint_status.id.org_id, [])
            blueprint_list.append(blueprint_status)

    def _skip_blueprint(self, blueprint: BlueprintStatusModel) -> bool:
        if self.blueprint_id is not None and blueprint.id.blueprint_id != self.blueprint_id:
            return True

        return False

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        blueprints = self._blueprints_by_org.get(org_config.github_id, [])

        if len(blueprints) > 0:
            github_id = org_config.github_id
            jsonnet_config = org_config.jsonnet_config
            await jsonnet_config.init_template()

            self._print_project_header(org_config, org_index, org_count)
            self.printer.level_up()

            try:
                org_file_name = jsonnet_config.org_config_file
                if not await self.check_config_file_exists(org_file_name):
                    return 1

                try:
                    organization = GitHubOrganization.load_from_file(github_id, org_file_name)
                except RuntimeError as ex:
                    self.printer.print_error(f"failed to load configuration: {ex!s}")
                    return 1

                try:
                    credentials = self.get_credentials(org_config, only_token=True)
                except RuntimeError as e:
                    self.printer.print_error(f"invalid credentials\n{e!s}")
                    return 1

                async with GitHubProvider(credentials) as provider:
                    rest_api = provider.rest_api

                    for blueprint in blueprints:
                        self.printer.print(f"Merging PR #{blueprint.remediation_pr}: ")

                        repo = organization.get_repository(blueprint.id.repo_name)
                        if repo is None:
                            self.printer.println("no repo found.")
                            continue

                        if repo.allow_merge_commit is True:
                            merge_method = "merge"
                        if repo.allow_squash_merge is True:
                            merge_method = "squash"
                        if repo.allow_rebase_merge is True:
                            merge_method = "rebase"

                        result = await rest_api.pull_request.merge_pull_request(
                            blueprint.id.org_id, blueprint.id.repo_name, f"{blueprint.remediation_pr}", merge_method
                        )

                        if result["merged"] is True:
                            self.printer.println("[green]merged[/].")
                        else:
                            self.printer.println(f"[red]failed[/]: {result['message']}.")

            finally:
                self.printer.level_down()

        return 0
